(*
  This file is distributed under the terms of the
  GNU Lesser General Public License version 2.1.
*)

(*
  A finite sets library for Coq.
  Copyright (C) 2006 Brian Aydemir and Aaron Bohannon

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*)

(*
  Some aspects of this file are inspired by the following file,
  originally distributed with Coq 8.1beta:

    (* Id: FSetInterface.v 8820 2006-05-15 11:44:05Z letouzey *)

  The modifications here are by:
    Brian Aydemir ( http://www.cis.upenn.edu/~baydemir/ )
    Aaron Bohannon ( http://www.cis.upenn.edu/~bohannon/ )

  The most recent modifcations were made on:
  $Date: 2006-08-24 14:38:25 -0400 (Thu, 24 Aug 2006) $
  $Id: ExtFset.v 766 2006-08-24 18:38:25Z baydemir $
*)

Require Import List.


(* *********************************************************************** *)
(** ** Finite sets with extensional equality *)

Record ExtFset (A : Set) : Type := mkExtFset {
  extFset : Set;

  (* Membership tests. *)
  In : A -> extFset -> Prop;
  In_dec : forall x E, {In x E} + {~ In x E};

  (* Equality. *)
  eq_carrier_dec : forall (x y : A), {x = y} + {x <> y};
  eq_extFset : forall E F, (forall x, In x E <-> In x F) -> E = F;

  (* Empty set. *)
  empty : extFset;
  empty_neg_intro : forall x, ~ In x empty;

  (* Adding an element to a set. *)
  add : A -> extFset -> extFset;
  add_intro_1 : forall x y E, x = y -> In x (add y E);
  add_intro_2 : forall x y E, In x E -> In x (add y E);
  add_elim : forall x y E, In x (add y E) -> x = y \/ In x E;

  (* Union. *)
  union : extFset -> extFset -> extFset;
  union_intro_1 : forall x E F, In x E -> In x (union E F);
  union_intro_2 : forall x E F, In x F -> In x (union E F);
  union_elim : forall x E F, In x (union E F) -> In x E \/ In x F;

  (* Intersection. *)
  intersection : extFset -> extFset -> extFset;
  intersection_intro :
    forall x E F, In x E -> In x F -> In x (intersection E F);
  intersection_elim_1 : forall x E F, In x (intersection E F) -> In x E;
  intersection_elim_2 : forall x E F, In x (intersection E F) -> In x F;

  (* Difference. *)
  diff : extFset -> extFset -> extFset;
  diff_intro : forall x E F, In x E -> ~ In x F -> In x (diff E F);
  diff_elim_1 : forall x E F, In x (diff E F) -> In x E;
  diff_elim_2 : forall x E F, In x (diff E F) -> ~ In x F;

  (* Map. *)
  map : (A -> A) -> extFset -> extFset;
  map_intro : forall x E f, In x E -> In (f x) (map f E);
  map_elim : forall y E f, In y (map f E) -> exists x, In x E /\ y = f x;

  (* The elements of a set as a (finite!) list. *)
  elements : extFset -> list A;
  elements_spec : forall x E, In x E <-> List.In x (elements E)
}.

Implicit Arguments extFset [A].
Implicit Arguments In [A e].
Implicit Arguments In_dec [A e].
Implicit Arguments eq_carrier_dec [A].

Implicit Arguments empty [A e].
Implicit Arguments empty_neg_intro [A e].

Implicit Arguments add [A e].
Implicit Arguments add_intro_1 [A e].
Implicit Arguments add_intro_2 [A e].
Implicit Arguments add_elim [A e].

Implicit Arguments union [A e].
Implicit Arguments union_intro_1 [A e].
Implicit Arguments union_intro_2 [A e].
Implicit Arguments union_elim [A e].

Implicit Arguments intersection [A e].
Implicit Arguments intersection_intro [A e].
Implicit Arguments intersection_elim_1 [A e].
Implicit Arguments intersection_elim_2 [A e].

Implicit Arguments diff [A e].
Implicit Arguments diff_intro [A e].
Implicit Arguments diff_elim_1 [A e].
Implicit Arguments diff_elim_2 [A e].

Implicit Arguments map [A e].
Implicit Arguments map_intro [A e].
Implicit Arguments map_elim [A e].

Implicit Arguments elements [A e].
Implicit Arguments elements_spec [A e].

Hint Resolve empty_neg_intro : sets.
Hint Resolve add_intro_1 add_intro_2 add_elim : sets.
Hint Resolve union_intro_1 union_intro_2 union_elim : sets.
Hint Resolve intersection_intro intersection_elim_1 intersection_elim_2 : sets.
Hint Resolve diff_intro diff_elim_1 diff_elim_2 : sets.
Hint Resolve map_intro map_elim : sets.
Hint Resolve elements_spec : sets.

Ltac destruct_empty :=
  match goal with
  | H : In (e := ?e) _ empty |- _ =>
    assert (fresh := @empty_neg_intro _ e _ H);
    contradiction
  end.

Ltac destruct_add H :=
  match goal with
  | _ : In (e := ?e) _ (add _ _) |- _ =>
    let H' := fresh in (
      assert (H' := @add_elim _ e _ _ _ H);
      clear H;
      inversion_clear H' as [H | H]
    )
  end.

Ltac destruct_union H :=
  match goal with
  | _ : In (e := ?e) _ (union _ _) |- _ =>
    let H' := fresh in (
      assert (H' := @union_elim _ e _ _ _ H);
      clear H;
      inversion_clear H' as [H | H]
    )
  end.

Ltac destruct_intersection H H1 H2 :=
  match goal with
  | _ : In (e := ?e) _ (intersection _ _) |- _ =>
    assert (H1 := @intersection_elim_1 _ e _ _ _ H);
    assert (H2 := @intersection_elim_2 _ e _ _ _ H);
    clear H
  end.

Ltac destruct_diff H H1 H2 :=
  match goal with
  | _ : In (e := ?e) _ (diff _ _) |- _ =>
    assert (H1 := @diff_elim_1 _ e _ _ _ H);
    assert (H2 := @diff_elim_2 _ e _ _ _ H);
    clear H
  end.

Ltac destruct_map H x H1 H2 :=
  match goal with
  | _ : In (e := ?e) _ (map _ _) |- _ =>
    case (@map_elim _ _ _ _ _ H); intros x [H1 H2];
    clear H
  end.


(* *********************************************************************** *)
(** ** Additional theorems *)

Section AdditionalTheorems.

Variable A : Set.
Variable ER : ExtFset A.

(** Empty. *)

Theorem empty_elim :
  forall (x : A) (E : extFset ER), In x (@empty _ ER) -> In x E.
Proof.
  intros x E H.
  absurd (In x (@empty _ ER)); auto with sets.
Qed.

(** Add. *)

Theorem add_elim_1 :
  forall (x y : A) (E : extFset ER), In x (add y E) -> ~ In x E -> x = y.
Proof.
  intros x y E H H'.
  destruct_add H; [trivial | contradiction].
Qed.

Theorem add_elim_2 :
  forall (x y : A) (E : extFset ER), In x (add y E) -> x <> y -> In x E.
Proof.
  intros x y E H H'.
  destruct_add H; [contradiction | trivial].
Qed.

Theorem add_neg_elim_1 :
  forall (x y : A) (E : extFset ER), ~ In x (add y E) -> x <> y.
Proof.
  auto with sets.
Qed.

Theorem add_neg_elim_2 :
  forall (x y : A) (E : extFset ER), ~ In x (add y E) -> ~ In x E.
Proof.
  auto with sets.
Qed.

Theorem add_neg_intro :
  forall (x y : A) (E : extFset ER), x <> y -> ~ In x E -> ~ In x (add y E).
Proof.
  intros x y E H1 H2 J.
  destruct_add J; contradiction.
Qed.

(** Union. *)

Theorem union_elim_1 :
  forall (x : A) (E F : extFset ER), In x (union E F) -> ~ In x F -> In x E.
Proof.
  intros x E F H J.
  destruct_union H; [trivial | contradiction].
Qed.

Theorem union_elim_2 :
  forall (x : A) (E F : extFset ER), In x (union E F) -> ~ In x E -> In x F.
Proof.
  intros x E F H J.
  destruct_union H; [contradiction | trivial].
Qed.

Theorem union_neg_elim_1 :
  forall (x : A) (E F : extFset ER), ~ In x (union E F) -> ~ In x E.
Proof.
  auto with sets.
Qed.

Theorem union_neg_elim_2 :
  forall (x : A) (E F : extFset ER), ~ In x (union E F) -> ~ In x F.
Proof.
  auto with sets.
Qed.

Theorem union_neg_intro :
  forall (x : A) (E F : extFset ER),
  ~ In x E -> ~ In x F -> ~ In x (union E F).
Proof.
  intros x E F H J L.
  destruct_union L; contradiction.
Qed.

(** Intersection. *)

Theorem intersection_neg_elim :
  forall (x : A) (E F : extFset ER),
  ~ In x (intersection E F) -> ~ In x E \/ ~ In x F.
Proof.
  intros x E F H.
  case (In_dec x E); auto.
  case (In_dec x F); auto.
  intros R J; assert (In x (intersection E F)); auto with sets.
Qed.

Theorem intersection_neg_elim_1 :
  forall (x : A) (E F : extFset ER),
  ~ In x (intersection E F) -> In x F -> ~ In x E.
Proof.
  auto with sets.
Qed.

Theorem intersection_neg_elim_2 :
  forall (x : A) (E F : extFset ER),
  ~ In x (intersection E F) -> In x E -> ~ In x F.
Proof.
  auto with sets.
Qed.

Theorem intersection_neg_intro_1 :
  forall (x : A) (E F : extFset ER), ~ In x E -> ~ In x (intersection E F).
Proof.
  eauto with sets.
Qed.

Theorem intersection_neg_intro_2 :
  forall (x : A) (E F : extFset ER), ~ In x F -> ~ In x (intersection E F).
Proof.
  eauto with sets.
Qed.

(** Diff. *)

Theorem diff_neg_elim :
  forall (x : A) (E F : extFset ER), ~ In x (diff E F) -> ~ In x E \/ In x F.
Proof.
  intros x E F H.
  case (In_dec x E); auto.
  case (In_dec x F); auto.
  intros R J; assert (In x (diff E F)); auto with sets.
Qed.

Theorem diff_neg_elim_1 :
  forall (x : A) (E F : extFset ER), ~ In x (diff E F) -> ~ In x F -> ~ In x F.
Proof.
  auto with sets.
Qed.

Theorem diff_neg_elim_2 :
  forall (x : A) (E F : extFset ER), ~ In x (diff E F) -> In x E -> In x F.
Proof.
  intros x E F H J.
  assert (Q := diff_neg_elim _ _ _ H).
  inversion_clear Q as [L1 | L2]; [contradiction | auto].
Qed.

Theorem diff_neg_intro_1 :
  forall (x : A) (E F : extFset ER), ~ In x E -> ~ In x (diff E F).
Proof.
  eauto with sets.
Qed.

Theorem diff_neg_intro_2 :
  forall (x : A) (E F : extFset ER), In x F -> ~ In x (diff E F).
Proof.
  intros x E F H J.
  destruct_diff J J1 J2; contradiction.
Qed.

(** Map. *)

Theorem map_compose :
  forall f g (F : extFset ER), map (fun x => f (g x)) F = map f (map g F).
Proof.
  intros f g F; apply eq_extFset; intros y; split; intros H.
  destruct_map H z K L; subst y.
    assert (J := map_intro _ _ g K); auto with sets.
  destruct_map H z K L; subst y.
    destruct_map K z' K' L'; subst z.
    change (f (g z')) with ((fun x => f (g x)) z'); auto with sets.
Qed.

(** Elements. *)

Theorem elements_empty :
  elements (@empty _ ER) = nil.
Proof.
  cut (forall x : A,
          In (e:=ER) x (@empty _ ER) <-> List.In x (@elements _ ER empty));
    auto with sets.
  induction (@elements _ ER empty) as [ | h t IH ]; auto.
  simpl; intros H.
  assert False; [idtac | intuition].
  pose (Q := proj2 (H h)).
  assert (~ In h (@empty _ ER)); auto with sets.
Qed.

End AdditionalTheorems.

Implicit Arguments empty_elim [A ER].

Implicit Arguments add_elim_1 [A ER].
Implicit Arguments add_elim_2 [A ER].
Implicit Arguments add_neg_elim_1 [A ER].
Implicit Arguments add_neg_elim_2 [A ER].
Implicit Arguments add_neg_intro [A ER].

Implicit Arguments union_elim_1 [A ER].
Implicit Arguments union_elim_2 [A ER].
Implicit Arguments union_neg_elim_1 [A ER].
Implicit Arguments union_neg_elim_2 [A ER].
Implicit Arguments union_neg_intro [A ER].

Implicit Arguments intersection_neg_elim [A ER].
Implicit Arguments intersection_neg_elim_1 [A ER].
Implicit Arguments intersection_neg_elim_2 [A ER].
Implicit Arguments intersection_neg_intro_1 [A ER].
Implicit Arguments intersection_neg_intro_2 [A ER].

Implicit Arguments diff_neg_elim [A ER].
Implicit Arguments diff_neg_elim_1 [A ER].
Implicit Arguments diff_neg_elim_2 [A ER].
Implicit Arguments diff_neg_intro_1 [A ER].
Implicit Arguments diff_neg_intro_2 [A ER].

Hint Resolve empty_elim : sets.
Hint Immediate add_elim_1 : sets.
Hint Resolve add_elim_2 : sets.
Hint Resolve add_neg_elim_1 add_neg_elim_2 add_neg_intro : sets.
Hint Resolve union_elim_1 union_elim_2 : sets.
Hint Resolve union_neg_elim_1 union_neg_elim_2 union_neg_intro : sets.
Hint Resolve intersection_neg_elim : sets.
Hint Resolve intersection_neg_elim_1 intersection_neg_elim_2 : sets.
Hint Resolve intersection_neg_intro_1 intersection_neg_intro_2 : sets.
Hint Resolve diff_neg_elim diff_neg_elim_1 diff_neg_elim_2 : sets.
Hint Resolve diff_neg_intro_1 diff_neg_intro_2 : sets.
Hint Resolve map_compose : sets.
Hint Resolve elements_empty : sets.

Hint Rewrite elements_empty : sets.

Ltac destruct_neg_add H H1 H2 :=
  match goal with
  | _ : ~ In (e := ?e) _ (add _ _) |- _ =>
    assert (H1 := @add_neg_elim_1 _ e _ _ _ H);
    assert (H2 := @add_neg_elim_2 _ e _ _ _ H);
    clear H
  end.

Ltac destruct_neg_union H H1 H2 :=
  match goal with
  | _ : ~ In (e := ?e) _ (union _ _) |- _ =>
    assert (H1 := @union_neg_elim_1 _ e _ _ _ H);
    assert (H2 := @union_neg_elim_2 _ e _ _ _ H);
    clear H
  end.

Ltac destruct_neg_intersection H :=
  match goal with
  | _ : ~ In (e := ?e) _ (intersection _ _) |- _ =>
    let H' := fresh in (
      assert (H' := @intersection_neg_elim _ e _ _ _ H);
      clear H;
      inversion_clear H' as [H | H]
    )
  end.

Ltac destruct_neg_diff H :=
  match goal with
  | _ : ~ In (e := ?e) _ (diff _ _) |- _ =>
    let H' := fresh in (
      assert (H' := @diff_neg_elim _ e _ _ _ H);
      clear H;
      inversion_clear H' as [H | H]
    )
  end.


(* *********************************************************************** *)
(** ** Singleton sets *)

Section Singleton.

Variable A : Set.
Variable ER : ExtFset A.

Definition singleton (x : A) : extFset ER := add x empty.

Theorem singleton_intro : forall (x y : A), x = y -> In x (singleton y).
Proof.
  intros x; unfold singleton; auto with sets.
Qed.

Theorem singleton_elim : forall (x y : A), In x (singleton y) -> x = y.
Proof.
  intros x y H.
  unfold singleton in H.
  destruct_add H; [trivial | destruct_empty].
Qed.

Theorem singleton_neg_elim : forall (x y : A), ~ In x (singleton y) -> x <> y.
Proof.
  intros x y H J.
  rewrite J in H; apply H; eapply singleton_intro; auto.
Qed.

Theorem singleton_neg_intro : forall (x y : A), x <> y -> ~ In x (singleton y).
Proof.
  intros x y H J.
  assert (x = y); [apply singleton_elim; auto | intuition].
Qed.

End Singleton.

Implicit Arguments singleton [A ER].
Implicit Arguments singleton_intro [A ER].
Implicit Arguments singleton_elim [A ER].
Implicit Arguments singleton_neg_elim [A ER].
Implicit Arguments singleton_neg_intro [A ER].

Hint Resolve singleton_intro : sets.
Hint Immediate singleton_elim : sets.
Hint Resolve singleton_neg_elim singleton_neg_intro : sets.

Ltac destruct_singleton H :=
  match goal with
  | _ : In (e := ?e) _ (singleton _) |- _ =>
    let H' := fresh in (
      rename H into H';
      assert (H := @singleton_elim _ e _ _ H');
      clear H'
    )
  end.

Ltac destruct_neg_singleton H :=
  match goal with
  | _ : ~ In (e := ?e) _ (singleton _) |- _ =>
    let H' := fresh in (
      rename H into H';
      assert (H := @singleton_neg_elim _ e _ _ H');
      clear H'
    )
  end.


(* *********************************************************************** *)
(** ** Removing an element from a set *)

Section Remove.

Variable A : Set.
Variable ER : ExtFset A.

Definition remove (x : A) (E : extFset ER) : extFset ER :=
  diff E (singleton x).

Theorem remove_intro :
  forall (x y : A) (E : extFset ER), x <> y -> In x E -> In x (remove y E).
Proof.
  unfold remove; intros x y E H J.
  auto with sets.
Qed.

Theorem remove_elim_1 :
  forall (x y : A) (E : extFset ER), In x (remove y E) -> x <> y.
Proof.
  unfold remove; intros x y E H.
  destruct_diff H J K; destruct_neg_singleton K; assumption.
Qed.

Theorem remove_elim_2 :
  forall (x y : A) (E : extFset ER), In x (remove y E) -> In x E.
Proof.
  unfold remove; intros x y E H.
  destruct_diff H J K; assumption.
Qed.

Lemma remove_neg_elim :
  forall (x y : A) (E : extFset ER),
  ~ In x (remove y E) -> x = y \/ ~ In x E.
Proof.
  unfold remove; intros x y E H.
  destruct_neg_diff H; [right | left; destruct_singleton H]; trivial.
Qed.

Lemma remove_neg_elim_strong_1 :
  forall (x y : A) (E : extFset ER),
  ~ In x (remove y E) -> x = y \/ (x <> y /\ ~ In x E).
Proof.
  intros x y E H.
  case (eq_carrier_dec ER x y); [auto | intro J].
  right; split; [trivial | idtac].
  case (remove_neg_elim _ _ _ H); intro K; [contradiction | trivial].
Qed.

Lemma remove_neg_elim_strong_2 :
  forall (x y : A) (E : extFset ER),
  ~ In x (remove y E) -> (x = y /\ In x E) \/ ~ In x E.
Proof.
  intros x y E H.
  case (In_dec x E); [intro J | auto].
  left; split; [idtac | trivial].
  case (remove_neg_elim _ _ _ H); intro K; [trivial | contradiction].
Qed.

Theorem remove_neg_intro_1 :
  forall (x y : A) (E : extFset ER), x = y -> ~ In x (remove y E).
Proof.
  unfold remove; intros x y E H J.
  destruct_diff J K L; destruct_neg_singleton L; contradiction.
Qed.

Theorem remove_neg_intro_2 :
  forall (x y : A) (E : extFset ER), ~ In x E -> ~ In x (remove y E).
Proof.
  unfold remove; intros x y E H J.
  destruct_diff J K L; destruct_neg_singleton L; contradiction.
Qed.

End Remove.

Implicit Arguments remove [A ER].
Implicit Arguments remove_intro [A ER].
Implicit Arguments remove_elim_1 [A ER].
Implicit Arguments remove_elim_2 [A ER].
Implicit Arguments remove_neg_elim [A ER].
Implicit Arguments remove_neg_elim_strong_1 [A ER].
Implicit Arguments remove_neg_elim_strong_2 [A ER].
Implicit Arguments remove_neg_intro_1 [A ER].
Implicit Arguments remove_neg_intro_2 [A ER].

Hint Resolve remove_intro remove_elim_1 remove_elim_2 : sets.
Hint Resolve remove_neg_elim remove_neg_intro_1 remove_neg_intro_2 : sets.
Hint Resolve remove_neg_elim_strong_1 remove_neg_elim_strong_2 : sets.

Ltac destruct_remove H H1 H2 :=
  match goal with
  | _ : In (e := ?e) _ (remove _ _) |- _ =>
    assert (H1 := @remove_elim_1 _ e _ _ _ H);
    assert (H2 := @remove_elim_2 _ e _ _ _ H);
    clear H
  end.

Ltac destruct_neg_remove H :=
  match goal with
  | _ : ~ In (e := ?e) _ (remove _ _) |- _ =>
    let H' := fresh in (
      assert (H' := @remove_neg_elim _ e _ _ _ H);
      clear H;
      inversion_clear H' as [H | H]
    )
  end.


(* *********************************************************************** *)
(** ** Equivalence theorems *)

Section EquivalenceTheorems.

Variable A : Set.
Variable ER : ExtFset A.

(** Add. *)

Lemma add_add_incl :
  forall (x y z: A) (E : extFset ER),
  In x (add y (add z E)) -> In x (add z (add y E)).
Proof.
  intros x y z E H.
  destruct_add H; [idtac | destruct_add H]; auto with sets.
Qed.

Theorem add_add :
  forall (x y : A) (E : extFset ER), add x (add y E) = add y (add x E).
Proof.
  intros x y E; apply eq_extFset; intro z; split; intro H.

    (* Direction 1 *)
    apply add_add_incl; trivial.

    (* Direction 2 *)
    apply add_add_incl; trivial.
Qed.

Theorem add_union :
  forall (x : A) (E F : extFset ER),
  add x (union E F) = union (add x E) (add x F).
Proof.
  intros x E F; apply eq_extFset; intro z; split; intro H.

    (* Direction 1 *)
    destruct_add H; [idtac | destruct_union H]; auto with sets.

    (* Direction 2 *)
    destruct_union H; destruct_add H; auto with sets.
Qed.

(** Union. *)

Theorem union_empty_1 :
  forall (E : extFset ER), union empty E = E.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1*)
    destruct_union H; auto with sets.

    (* Direction 2*)
    auto with sets.
Qed.

Theorem union_empty_2 :
  forall (E : extFset ER), union E empty = E.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1*)
    destruct_union H; auto with sets.

    (* Direction 2*)
    auto with sets.
Qed.

Theorem union_commute :
  forall (E F : extFset ER), union E F = union F E.
Proof.
  intros E F; apply eq_extFset; intro x; split; intro H.

    (* Direction 1 *)
    destruct_union H; auto with sets.

    (* Direction 2 *)
    destruct_union H; auto with sets.
Qed.

Theorem union_union_distrib :
  forall (E F G : extFset ER),
  union (union E F) G = union (union E G) (union F G).
Proof.
  intros E F G; apply eq_extFset; intro x.
  split; intro H.

    (* Direction 1 *)
    destruct_union H; auto with sets.
    destruct_union H; auto with sets.

    (* Direction 2 *)
    destruct_union H; destruct_union H; auto with sets.
Qed.

(** Intersection. *)

Theorem intersection_empty_1 :
  forall (E : extFset ER), intersection empty E = empty.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1 *)
    eauto with sets.

    (* Direction 2 *)
    eauto with sets.
Qed.

Theorem intersection_empty_2 :
  forall (E : extFset ER), intersection E empty = empty.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1 *)
    eauto with sets.

    (* Direction 2 *)
    eauto with sets.
Qed.

(** Diff. *)

Theorem diff_empty_1 :
  forall (E : extFset ER), diff E empty = E.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1 *)
    destruct_diff H J K; auto with sets.

    (* Direction 2 *)
    auto with sets.
Qed.

Theorem diff_empty_2 :
  forall (E : extFset ER), diff empty E = empty.
Proof.
  intros E; apply eq_extFset; intro x; split; intro H.

    (* Direction 1 *)
    eauto with sets.

    (* Direction 2 *)
    eauto with sets.
Qed.

(** Remove. *)

Theorem remove_empty :
  forall (x : A), remove x (@empty _ ER) = empty.
Proof.
  intros x; apply eq_extFset.
  intros y; split; intro H.

  (* Direction 1 *)
  destruct_remove H J K; trivial.

  (* Direction 2 *)
  auto with sets.
Qed.

Theorem remove_union :
  forall (x : A) (E F : extFset ER),
  remove x (union E F) = union (remove x E) (remove x F).
Proof.
  intros x E F; apply eq_extFset.
  intros y; split; intros H.

  (* Direction 1 *)
  destruct_remove H J K; destruct_union K; auto with sets.

  (* Direction 2 *)
  destruct_union H; destruct_remove H J K; auto with sets.
Qed.

Theorem remove_singleton_1 :
  forall (x y : A), x = y -> remove x (singleton y) = @empty _ ER.
Proof.
  intros x y H; rewrite H; apply eq_extFset.
  intros z; split; intros J.

  (* Direction 1 *)
  destruct_remove J K L; destruct_singleton L; contradiction.

  (* Direction 2 *)
  destruct_empty.
Qed.

Theorem remove_singleton_2 :
  forall (x y : A),
  x <> y -> remove (ER := ER) x (singleton y) = (singleton y).
Proof.
  intros x y H; apply eq_extFset.
  intros z; split; intros J.

  (* Direction 1 *)
  destruct_remove J K L; trivial.

  (* Direction 2 *)
  destruct_singleton J; rewrite J; auto with sets.
Qed.

End EquivalenceTheorems.

Implicit Arguments add_add_incl [A ER].
Implicit Arguments add_add [A ER].
Implicit Arguments add_union [A ER].
Implicit Arguments union_empty_1 [A ER].
Implicit Arguments union_empty_2 [A ER].
Implicit Arguments union_union_distrib [A ER].
Implicit Arguments intersection_empty_1 [A ER].
Implicit Arguments intersection_empty_2 [A ER].
Implicit Arguments diff_empty_1 [A ER].
Implicit Arguments diff_empty_2 [A ER].
Implicit Arguments remove_empty [A ER].
Implicit Arguments remove_union [A ER].
Implicit Arguments remove_singleton_1 [A ER].
Implicit Arguments remove_singleton_2 [A ER].

Hint Immediate add_add_incl : sets.
Hint Resolve add_add add_union : sets.
Hint Resolve union_empty_1 union_empty_2 : sets.
Hint Immediate union_commute : sets.
Hint Resolve union_union_distrib : sets.
Hint Resolve intersection_empty_1 intersection_empty_2 : sets.
Hint Resolve diff_empty_1 diff_empty_2 : sets.
Hint Resolve remove_empty remove_union : sets.
Hint Resolve remove_singleton_1 remove_singleton_2 : sets.

Hint Rewrite add_union : sets.
Hint Rewrite union_empty_1 union_empty_2 : sets.
Hint Rewrite intersection_empty_1 intersection_empty_2 : sets.
Hint Rewrite diff_empty_1 diff_empty_2 : sets.
Hint Rewrite remove_empty remove_union : sets.
Hint Rewrite remove_singleton_1 using (auto; fail) : sets.
Hint Rewrite remove_singleton_2 using (auto; fail) : sets.


(* *********************************************************************** *)
(** ** Predicates on sets. *)

Section Predicates.

Variable A : Set.
Variable ER : ExtFset A.

Definition Subset (E F : extFset ER) := forall x, In x E -> In x F.

Theorem Subset_refl :
  forall (E : extFset ER), Subset E E.
Proof.
  unfold Subset; auto.
Qed.

Theorem Subset_trans :
  forall (E F G : extFset ER),
  Subset E F -> Subset F G -> Subset E G.
Proof.
  unfold Subset; intuition.
Qed.

Theorem Subset_empty :
  forall (E : extFset ER), Subset empty E.
Proof.
  unfold Subset; eauto with sets.
Qed.

Theorem Subset_singleton_l :
  forall (E : extFset ER) (x : A), In x E -> Subset (singleton x) E.
Proof.
  unfold Subset; intros E x H x' H'.
  destruct_singleton H'; subst x'; assumption.
Qed.

Theorem Subset_union_l :
  forall (E F G : extFset ER), Subset E G -> Subset F G -> Subset (union E F) G.
Proof.
  unfold Subset; intros E F G H J x K.
  destruct_union K; auto.
Qed.

Theorem Subset_union_r1 :
  forall (E F : extFset ER), Subset E (union E F).
Proof.
  unfold Subset; auto with sets.
Qed.

Theorem Subset_union_r2 :
  forall (E F : extFset ER), Subset F (union E F).
Proof.
  unfold Subset; auto with sets.
Qed.

Theorem Subset_union_union :
  forall (A B C D : extFset ER),
  Subset A C -> Subset B D -> Subset (union A B) (union C D).
Proof.
  unfold Subset; intros A' B C D H J x K.
  destruct_union K; auto with sets.
Qed.

Theorem Subset_add :
  forall (E F : extFset ER) (x : A),
  Subset E F -> Subset (add x E) (add x F).
Proof.
  unfold Subset; intros E F x H x' H'.
  destruct_add H'.
    subst x'; auto with sets.
    auto with sets.
Qed.

Theorem Subset_add_r :
  forall (E F G : extFset ER) (x : A),
  Subset F G -> Subset E (add x F) -> Subset E (add x G).
Proof.
  intros E F G x H J.
  assert (Subset (add x F) (add x G)); auto using Subset_add.
  eauto using Subset_trans.
Qed.

Theorem Subset_remove_l_to_add :
  forall (E F : extFset ER) (x : A),
  Subset (remove x E) F -> Subset E (add x F).
Proof.
  unfold Subset; intros E F x H x' H'.
  case (eq_carrier_dec ER x' x); auto with sets.
Qed.

Theorem Subset_add_l_to_remove :
  forall (E F : extFset ER) (x : A),
  Subset E (add x F) -> Subset (remove x E) F.
Proof.
  unfold Subset; intros E F x H x' H'.
  destruct_remove H' J K; case (eq_carrier_dec ER x' x); intros EQ; try subst.
  intuition.
  assert (L := H _ K); destruct_add L; intuition.
Qed.

End Predicates.

Implicit Arguments Subset [A ER].

Hint Resolve Subset_refl : sets.
Hint Resolve Subset_empty Subset_singleton_l : sets.
Hint Resolve Subset_union_l : sets.
Hint Resolve Subset_union_r1 Subset_union_r2 Subset_union_union : sets.
Hint Resolve Subset_add Subset_add_r : sets.
Hint Resolve Subset_remove_l_to_add Subset_add_l_to_remove : sets.
