Require Import List.
Require Import Atoms.
Require Import Axioms.
Require Import Permutations.


(**
  A [PsetT] record is the evidence that we can define permutations
  over a type [X].  Both [perm_id] and [perm_compose] are standard
  axioms of group actions (in this case, of permutations acting on [X]).
*)
Record PsetT (A : AtomT) (X : Set) : Set := mkPset {
  perm : permt A -> X -> X;

  perm_id : forall x, perm [] x = x;

  perm_compose :
    forall p q x, perm (p :++ q) x = perm p (perm q x)
}.

Implicit Arguments mkPset [A X].
Implicit Arguments perm [A X].
Implicit Arguments perm_id [A X].
Implicit Arguments perm_compose [A X].

Hint Resolve perm_id perm_compose : nominal.
Hint Rewrite perm_id perm_compose : nominal.


(* ********************************************************************** *)
(** ** Basic properties about psets *)

Section BasicPsetTheorems.
  Variable A : AtomT.
  Variable X : Set.
  Variable P : PsetT A X.

  Lemma perm_eq : forall p q x, p = q -> perm P p x = perm P q x.
  Proof.
    congruence.
  Qed.

  Hint Resolve perm_eq : nominal.

  Lemma perm_inv : forall p x, perm P (-p) (perm P p x) = x.
  Proof.
    intros p x.
    pattern x at 2; rewrite <- (perm_id P x).
    rewrite <- perm_compose.
    auto with nominal.
  Qed.

  Lemma perm_inv' : forall p x, perm P (p) (perm P (-p) x) = x.
  Proof.
    intros p x.
    pattern x at 2; rewrite <- (perm_id P x).
    rewrite <- perm_compose.
    auto with nominal.
  Qed.

  Lemma perm_injective :
    forall p x y, perm P p x = perm P p y -> x = y. 
  Proof. 
    intros p x y H.
    rewrite <- (perm_inv p x); rewrite <- (perm_inv p y).
    rewrite H.
    reflexivity.
  Qed.

  Lemma perm_swap_same :
    forall a x, perm P [(a, a)] x = x.
  Proof.
    intros a x; autorewrite with nominal; trivial.
  Qed.

  Lemma perm_swap_invol :
    forall ab x, perm P [ab] (perm P [ab] x) = x.
  Proof.
    intros ab x.
    pattern x at 2; rewrite <- (perm_id P x).
    rewrite <- perm_compose.
    auto with nominal.
  Qed.

  Lemma perm_swap_move :
    forall ab x y, x = perm P [ab] y -> perm P [ab] x = y.
  Proof.
    intros ab x y H.
    assert (Q : perm P [ab] x = perm P [ab] (perm P [ab] y)); try congruence.
    rewrite perm_swap_invol in Q.
    assumption.
  Qed.

  Lemma perm_swap_move' :
    forall ab x y, perm P [ab] x = y -> x = perm P [ab] y.
  Proof.
    intros ab x y H.
    assert (Q : perm P [ab] (perm P [ab] x) = perm P [ab] y); try congruence.
    rewrite perm_swap_invol in Q.
    assumption.
  Qed.

  Lemma perm_swap_switch :
    forall a b x, perm P [(a, b)] x = perm P [(b, a)] x.
  Proof.
    intros a b x; auto with nominal.
  Qed.

End BasicPsetTheorems.

Hint Resolve perm_inv perm_inv' : nominal.
Hint Rewrite perm_inv perm_inv' : nominal.
Hint Immediate perm_injective : nominal.
Hint Resolve perm_swap_same perm_swap_invol : nominal.
Hint Rewrite perm_swap_same perm_swap_invol : nominal.
Hint Immediate perm_swap_move perm_swap_move' : nominal.
Hint Resolve perm_swap_switch : nominal.


(* ********************************************************************** *)
(** ** Support *)

(**
  We define what it means for a finite set of atoms to support an
  element of a pset.
*)
Section Support.
  Variable A : AtomT.
  Variable X : Set.
  Variable P : PsetT A X.

  Definition supports (F : aset A) (x : X) :=
    forall a b, ~ In a F -> ~ In b F -> perm P [(a, b)] x = x.

  Lemma swap_non_supports :
    forall F x a b, supports F x -> ~ In a F -> ~ In b F ->
    perm P [(a, b)] x = x.
  Proof.
    unfold supports; auto.
  Qed.

  Lemma Subset_supports :
    forall (E F : aset A) (x : X),
    supports E x -> Subset E F -> supports F x.
  Proof.
    unfold supports; intros E F x H J c d C D.
    rewrite H; intuition.
  Qed.

  Lemma intersection_supports :
    forall (E F : aset A) (x : X),
    supports E x -> supports F x -> supports (intersection E F) x.
  Proof.
    unfold supports; intros E F x H J a a' AA AA'.
    case (atom_eqdec A a a'); intro T.

  (* Case: a = a' *)
  rewrite T; rewrite perm_swap_same; auto.

  (* Case: a <> a' *)
  case (atom_infinite A (add a' (add a (union E F)))); intros a'' M.
  assert (M1 : ~ In a'' E).
    eapply union_neg_elim_1; do 2 eapply add_neg_elim_2; eauto with sets.
  assert (M2 : ~ In a'' F).
    eapply union_neg_elim_2; do 2 eapply add_neg_elim_2; eauto with sets.
  assert (M3 : a'' <> a).
    apply add_neg_elim_1 with (E := union E F).
    apply add_neg_elim_2 with (y := a'); intuition.
  assert (M4 : a'' <> a').
    eapply add_neg_elim_1 with (E := add a (union E F)); intuition.

  assert (Q : perm P [(a, a')] x =
              perm P [(a, a'')] (perm P [(a', a'')] (perm P [(a, a'')] x))).
  repeat rewrite <- perm_compose; apply perm_eq; apply perma_eq; intros z.
  autorewrite with nominal.
  rewrite swapa_distrib; swapa_simpl; autorewrite with nominal.
  rewrite swapa_switch; reflexivity.

  rewrite Q.
  assert (Q1 : ~ In a E \/ ~ In a F); auto with sets.
  assert (Q2 : ~ In a' E \/ ~ In a' F); auto with sets.
  inversion_clear Q1 as [U | U]; inversion_clear Q2 as [V | V].

  repeat rewrite H; trivial.

  replace (perm P [(a, a'')] x) with x; [idtac | rewrite H; trivial].
  replace (perm P [(a', a'')] x) with x; [idtac | rewrite J; trivial].
  replace (perm P [(a, a'')] x) with x; [auto | rewrite H; trivial].

  replace (perm P [(a, a'')] x) with x; [idtac | rewrite J; trivial].
  replace (perm P [(a', a'')] x) with x; [idtac | rewrite H; trivial].
  replace (perm P [(a, a'')] x) with x; [auto | rewrite J; trivial].

  repeat rewrite J; trivial.
Qed.

End Support.

Implicit Arguments supports [A X].
Implicit Arguments swap_non_supports [A X].

Hint Resolve swap_non_supports : nominal.
Hint Resolve Subset_supports intersection_supports : nominal.


(* ********************************************************************** *)
(** ** Identity permutation *)

Section IdPermutations.
  Variable A : AtomT.
  Variable X : Set.

  Definition mkIdPset : PsetT A X :=
    mkPset (fun _ x => x) (fun _ => refl_equal _) (fun _ _ _ => refl_equal _).

End IdPermutations.


(* ********************************************************************** *)
(** ** Permutations on atoms *)

Section AtomPermutations.
  Variable A : AtomT.

  Definition mkAtomPset :=
    mkPset (@perma A) (perma_id A) (perma_compose A).

  Lemma supports_atom :
    forall a : A, supports mkAtomPset (singleton a) a.
  Proof.
    unfold supports; intros a c d C D; simpl.
    autorewrite with nominal.
    destruct_neg_singleton C; destruct_neg_singleton D.
    swapa_simpl.
    reflexivity.
  Qed.

  Lemma perm_atom : forall p a, perm mkAtomPset p a = p @ a.
  Proof.
    reflexivity.
  Qed.

End AtomPermutations.

Implicit Arguments perm_atom [A].

Hint Resolve supports_atom : nominal.


(* ********************************************************************** *)
(** ** Permutations on pairs *)

Section PairPermutations.
  Variable A : AtomT.
  Variables X Y : Set.
  Variables (XP : PsetT A X) (YP : PsetT A Y).

  Definition perm_on_pair (p : permt A) (xy : X * Y) :=
    let (x, y) := xy in (perm XP p x, perm YP p y).

  Definition mkPairPset : PsetT A (X * Y).
    refine (mkPset perm_on_pair _ _).
    induction x; simpl; autorewrite with nominal; reflexivity.
    intros p q; induction x; simpl; autorewrite with nominal; reflexivity.
  Defined.

  Lemma perm_pair :
    forall p x y, perm mkPairPset p (x, y) = (perm XP p x, perm YP p y).
  Proof.
    reflexivity.
  Qed.

  Lemma supports_pair :
    forall x y F G,
    supports XP F x ->
    supports YP G y ->
    supports mkPairPset (union F G) (x, y).
  Proof.
    unfold supports; intros x y F G H J c d C D.
    rewrite perm_pair.
    rewrite H; auto with sets.
    rewrite J; auto with sets.
  Qed.

End PairPermutations.

Implicit Arguments mkPairPset [A X Y].

Notation "X ^* Y" := (mkPairPset X Y) (at level 50, left associativity).

Hint Resolve perm_pair supports_pair : nominal.
Hint Rewrite perm_pair : nominal.


(* ********************************************************************** *)
(** ** Permutations on lists *)

Section ListPermutations.
  Variable A : AtomT.
  Variable X : Set.
  Variable P : PsetT A X.

  Fixpoint perm_on_list (p : permt A) (xs : list X) {struct xs} : list X :=
    match xs with
      | nil => nil
      | y :: ys => (perm P p y) :: (perm_on_list p ys)
    end.

  Definition mkListPset : PsetT A (list X).
    refine (mkPset perm_on_list _ _).
    induction x; simpl; autorewrite with nominal; congruence.
    induction x; simpl; autorewrite with nominal; congruence.
  Defined.

  Lemma perm_list_nil : forall p, perm mkListPset p nil = nil.
  Proof.
    reflexivity.
  Qed.

  Lemma perm_list_cons :
    forall p x xs,
    perm mkListPset p (x :: xs) = perm P p x :: perm mkListPset p xs.
  Proof.
    reflexivity.
  Qed.

  Lemma perm_list_In :
    forall xs x p,
    List.In x xs -> List.In (perm P p x) (perm mkListPset p xs).
  Proof.
    induction xs as [ | h t IHt ]; simpl; trivial.
    intros x p [H | H]; try subst h; auto.
  Qed.

  Lemma perm_list_In' :
    forall xs x p,
    List.In (perm P p x) (perm mkListPset p xs) -> List.In x xs.
  Proof.
    intros xs x p H.
    assert (Q := perm_list_In _ _ (- p) H).
    repeat rewrite perm_inv in Q; trivial.
  Qed.

  Lemma perm_list_not_In :
    forall xs x p,
    ~ List.In x xs -> ~ List.In (perm P p x) (perm mkListPset p xs).
  Proof.
    intros xs x p H J.
    assert (List.In x xs) by eauto using perm_list_In'.
    contradiction.
  Qed.

  Lemma perm_list_not_In' :
    forall xs x p,
    ~ List.In (perm P p x) (perm mkListPset p xs) -> ~ List.In x xs.
  Proof.
    intros xs x p H J.
    auto using perm_list_In.
  Qed.

  Lemma supports_list_nil : supports mkListPset empty nil.
  Proof.
    unfold supports; simpl; intros.
    reflexivity.
  Qed.

  Lemma supports_list_cons :
    forall F G x xs, supports P F x -> supports mkListPset G xs ->
    supports mkListPset (union F G) (x :: xs).
  Proof.
    intros F G x xs H J; unfold supports; intros c d C D; simpl.
    rewrite (swap_non_supports P F x); auto with sets.
    rewrite (swap_non_supports mkListPset G xs); auto with sets.
  Qed.

End ListPermutations.

Implicit Arguments mkListPset [A X].

Hint Resolve perm_list_nil perm_list_cons : nominal.
Hint Rewrite perm_list_nil perm_list_cons : nominal.
Hint Resolve perm_list_In perm_list_In' : nominal.
Hint Resolve perm_list_not_In perm_list_not_In' : nominal.
Hint Resolve supports_list_nil supports_list_cons : nominal.


(* ********************************************************************** *)
(** ** Permutations on finite sets *)

Section FiniteSetPermutations.
  Variable A : AtomT.
  Variable X : Set.
  Variable E : ExtFset X.
  Variable P : PsetT A X.

  Definition perm_on_extFset (p : permt A) (F : extFset E) :=
    map (fun x => perm P p x) F.

  Lemma perm_on_extFset_In_raw :
    forall x F p, In x F -> In (perm P p x) (perm_on_extFset p F).
  Proof.
    intros x F p H; unfold perm_on_extFset.
    exact (map_intro x F (fun x => perm P p x) H).
  Qed.

  Hint Resolve perm_on_extFset_In_raw : nominal.

  Definition mkExtFsetPset : PsetT A (extFset E).
    refine (mkPset perm_on_extFset _ _).
    intros x; apply eq_extFset; intros y; split; intros H.
      unfold perm_on_extFset in H; destruct_map H z K L.
        rewrite perm_id in L; subst y; trivial.
      assert (J := map_intro _ _ (fun x => perm P [] x) H).
        simpl in J; rewrite perm_id in J; trivial.
    intros p q x; unfold perm_on_extFset.
      replace (fun x => perm P (p :++ q) x)
         with (fun x => perm P p (perm P q x)).
      apply (map_compose _ E (fun x => perm P p x) (fun x => perm P q x)).
      apply fun_extensionality; auto with nominal.
  Defined.

  Lemma perm_extFset_In :
    forall x F p, In x F -> In (perm P p x) (perm mkExtFsetPset p F).
  Proof.
    auto with nominal.
  Qed.

  Lemma perm_extFset_In' :
    forall x F p,  In (perm P p x) (perm mkExtFsetPset p F) -> In x F.
  Proof.
    intros x F p H.
    assert (J := perm_extFset_In _ _ (- p) H).
    autorewrite with nominal in J.
    assumption.
  Qed.

  Lemma perm_extFset_neg_In :
    forall x F p, ~ In x F -> ~ In (perm P p x) (perm mkExtFsetPset p F).
  Proof.
    intros x F p H J.
    assert (K := perm_extFset_In' _ _ _ J); firstorder.
  Qed.

  Lemma perm_extFset_neg_In' :
    forall x F p,  ~ In (perm P p x) (perm mkExtFsetPset p F) -> ~ In x F.
  Proof.
    intuition (auto using perm_extFset_In).
  Qed.

  Lemma perm_extFset_empty : forall p, perm mkExtFsetPset p empty = empty.
  Proof.
    intros p; apply eq_extFset; intros x; split; intros H.
    assert (J := perm_extFset_In _ _ (-p) H).
      autorewrite with nominal in J.
      destruct_empty.
    destruct_empty.
  Qed.

  Lemma perm_extFset_add :
    forall x F p,
    perm mkExtFsetPset p (add x F) = add (perm P p x) (perm mkExtFsetPset p F).
  Proof.
    intros x F p; apply eq_extFset; intros y; split; intros H.
    assert (J := perm_extFset_In _ _ (-p) H).
      autorewrite with nominal in J.
      destruct_add J.
        subst x; autorewrite with nominal.
          auto with sets.
        assert (K := perm_extFset_In _ _ p J).
          autorewrite with nominal in K.
          auto with sets.
    destruct_add H.
      subst y; auto with nominal sets.
      replace y with (perm P p (perm P (-p) y)); auto with nominal.
        assert (J := perm_extFset_In _ _ (-p) H).
        autorewrite with nominal in J.
        auto with nominal sets.
  Qed.

  Lemma perm_extFset_singleton :
    forall x p,
    perm mkExtFsetPset p (singleton x) = singleton (perm P p x).
  Proof.
    intros x p; unfold singleton.
    rewrite perm_extFset_add; rewrite perm_extFset_empty.
    reflexivity.
  Qed.

  Lemma perm_extFset_union :
    forall p G F,
    perm mkExtFsetPset p (union G F) =
    union (perm mkExtFsetPset p G) (perm mkExtFsetPset p F).
  Proof.
    intros p G F; apply eq_extFset; intros x.
    replace x with (perm P p (perm P (-p) x)); auto with nominal.
    split; intros H.
    assert (J := perm_extFset_In' _ _ _ H).
      destruct_union J; auto using perm_extFset_In with sets.
    apply perm_extFset_In; destruct_union H.
      assert (J := perm_extFset_In' _ _ _ H); auto with sets.
      assert (J := perm_extFset_In' _ _ _ H); auto with sets.
  Qed.

  Lemma perm_extFset_diff :
    forall p G F,
    perm mkExtFsetPset p (diff G F) =
    diff (perm mkExtFsetPset p G) (perm mkExtFsetPset p F).
  Proof.
    intros p G F; apply eq_extFset; intros x.
    replace x with (perm P p (perm P (-p) x)); auto with nominal.
    split; intros H.
    assert (J := perm_extFset_In' _ _ _ H).
      destruct_diff J K L;
        auto using perm_extFset_In, perm_extFset_neg_In with sets.
    apply perm_extFset_In; destruct_diff H J K.
      assert (L := perm_extFset_In' _ _ _ J).
      assert (M := perm_extFset_neg_In' _ _ _ K).
      auto with sets.
  Qed.

  Lemma perm_extFset_remove :
    forall p x F,
    perm mkExtFsetPset p (remove x F) =
    remove (perm P p x) (perm mkExtFsetPset p F).
  Proof.
    unfold remove; intros p x F.
    rewrite <- perm_extFset_singleton.
    apply perm_extFset_diff.
  Qed.

  Lemma perm_extFset_Subset :
    forall p G F, Subset G F ->
    Subset (perm mkExtFsetPset p G) (perm mkExtFsetPset p F).
  Proof.
    unfold Subset; intros p G F H x.
    replace x with (perm P p (perm P (-p) x)); auto with nominal.
    intros J.
    assert (K := perm_extFset_In' _ _ _ J).
    exact (perm_extFset_In _ _ _ (H _ K)).
  Qed.

  Lemma swap_not_in_set :
    forall (a c d : A) (F : aset A),
    ~ In a F -> ~ In c F -> ~ In d F -> ~ In ([(c,d)] @ a) F.
  Proof.
    intros a c d F H J K.
    autorewrite with nominal.
    swapa_case M N c d a; trivial.
  Qed.

  Lemma swap_not_in_set' :
    forall (a c d : A) (F : aset A),
    ~ In a F -> ~ In c F -> ~ In d F -> ~ In (perm (mkAtomPset A) [(c,d)] a) F.
  Proof.
    intros a c d F H J K.
    rewrite perm_atom.
    apply swap_not_in_set; trivial.
  Qed.

End FiniteSetPermutations.

Implicit Arguments mkExtFsetPset [A X].
Implicit Arguments perm_extFset_In [A X E x F].
Implicit Arguments perm_extFset_In' [A X E x F].
Implicit Arguments perm_extFset_neg_In [A X E x F].
Implicit Arguments perm_extFset_neg_In' [A X E x F].

Hint Resolve perm_extFset_In : nominal.
Hint Resolve perm_extFset_neg_In : nominal.
Hint Immediate perm_extFset_In' : nominal.
Hint Immediate perm_extFset_neg_In' : nominal.
Hint Resolve perm_extFset_empty perm_extFset_add : nominal.
Hint Rewrite perm_extFset_empty perm_extFset_add : nominal.
Hint Resolve perm_extFset_singleton perm_extFset_union : nominal.
Hint Rewrite perm_extFset_singleton perm_extFset_union : nominal.
Hint Resolve perm_extFset_diff perm_extFset_remove : nominal.
Hint Rewrite perm_extFset_diff perm_extFset_remove : nominal.
Hint Resolve perm_extFset_Subset : nominal.
Hint Resolve swap_not_in_set swap_not_in_set' : nominal.


(* ********************************************************************** *)
(** ** Freshness as a predicate *)

Section Fresh.
  Variable A : AtomT.
  Variable X : Set.
  Variable P : PsetT A X.

  Lemma perm_supports :
    forall F x p,
    supports P F x ->
    supports P (perm (mkExtFsetPset (asetR A) (mkAtomPset A)) p F) (perm P p x).
  Proof.
    unfold supports; intros F x p H c d C D.
    apply (perm_injective _ _ P (-p)).
    rewrite <- perm_compose; rewrite perma_swap_commute.
    autorewrite with nominal.
    assert (Q := perm_extFset_neg_In (mkAtomPset A) (-p) C).
    assert (R := perm_extFset_neg_In (mkAtomPset A) (-p) D).
    autorewrite with nominal in Q.
    autorewrite with nominal in R.
    rewrite perm_atom in * |-.
    rewrite H; trivial.
  Qed.

  Lemma perm_supports' :
    forall F x p,
    supports P (perm (mkExtFsetPset (asetR A) (mkAtomPset A)) p F) (perm P p x) ->
    supports P F x.
  Proof.
    intros a x p H.
    assert (J := perm_supports _ _ (-p) H).
    autorewrite with nominal in J.
    assumption.
  Qed.

  Definition freshP (a : A) (x : X) :=
    exists F : aset A, ~ In a F /\ supports P F x.

  Lemma freshP_from_supports :
    forall a x F, ~ In a F -> supports P F x -> freshP a x.
  Proof.
    unfold freshP; intros a x F H J; exists F; auto.
  Qed.

  Lemma perm_freshP :
    forall a x p, freshP a x -> freshP (perm (mkAtomPset A) p a) (perm P p x).
  Proof.
    unfold freshP; intros a x p H.
    inversion_clear H as [F [Q R]].
    exists (perm (mkExtFsetPset (asetR A) (mkAtomPset A)) p F).
    auto using perm_supports with nominal.
  Qed.

  Lemma perm_freshP' :
    forall a x p, freshP (perm (mkAtomPset A) p a) (perm P p x) -> freshP a x.
  Proof.
    intros a x p H.
    assert (J := perm_freshP _ _ (-p) H).
    autorewrite with nominal in J.
    assumption.
  Qed.

  Lemma swap_freshP_atoms :
    forall a b x, freshP a x -> freshP b x -> perm P [(a,b)] x = x.
  Proof.
    intros a b x [F1 [H J]] [F2 [L K]].
    rewrite swap_non_supports with (F := intersection F1 F2);
      auto with nominal sets.
  Qed.

End Fresh.

Implicit Arguments freshP [A X].

Hint Resolve perm_supports : nominal.
Hint Immediate perm_supports' : nominal.
Hint Resolve perm_freshP : nominal.
Hint Immediate perm_freshP' : nominal.


Section FreshFacts.
  Variable A : AtomT.

  Lemma freshP_atom :
    forall a b, a <> b -> freshP (mkAtomPset A) a b.
  Proof.
    intros a b H; unfold freshP.
    exists (@singleton _ (asetR A) b); split.
    intros J; destruct_singleton J; contradiction.
    auto with nominal.
  Qed.

End FreshFacts.

Hint Resolve freshP_from_supports : nominal.
Hint Resolve freshP_atom : nominal.


(* ********************************************************************** *)
(** ** Permutations on functions *)

Section FunctionPermutations.
  Variable A : AtomT.
  Variables X Y : Set.
  Variables (XP : PsetT A X) (YP : PsetT A Y).

  Definition perm_on_fun (p : permt A) (f : X -> Y) :=
    fun x => perm YP p (f (perm XP (-p) x)).

  Definition mkFunPset : PsetT A (X -> Y).
    refine (mkPset perm_on_fun _ _).
    intros f; apply fun_extensionality; intros x; unfold perm_on_fun.
      autorewrite with nominal; reflexivity.
    intros p q f; apply fun_extensionality; intros x; unfold perm_on_fun.
      autorewrite with nominal; reflexivity.
  Defined.

  Lemma perm_fun :
    forall f p, perm mkFunPset p f = fun x => perm YP p (f (perm XP (-p) x)).
  Proof.
    reflexivity.
  Qed.

  Lemma perm_app :
    forall f x p, perm YP p (f x) = (perm mkFunPset p f) (perm XP p x).
  Proof.
    intros f x p; rewrite perm_fun.
    autorewrite with nominal.
    reflexivity.
  Qed.

  Lemma supports_fun :
    forall (F : aset A) (f : X -> Y),
    (forall a b x, ~ In a F -> ~ In b F ->
      perm YP [(a, b)] (f x) = f (perm XP [(a, b)] x)) ->
    supports mkFunPset F f.
  Proof.
    unfold supports; intros F f H; intros c d C D.
    apply fun_extensionality; intros x; rewrite perm_fun.
    autorewrite with nominal.
    rewrite <- H; trivial.
    rewrite perm_swap_invol; reflexivity.
  Qed.

  Lemma supports_application :
    forall f x F G,
    supports mkFunPset F f ->
    supports XP G x ->
    supports YP (union F G) (f x).
  Proof.
    unfold supports; intros f x F G H J c d C D.
    rewrite perm_app.
    rewrite swap_non_supports with (F := F) (x := f); auto with sets.
    rewrite swap_non_supports with (F := G) (x := x); auto with sets.
  Qed.

  Lemma freshP_application :
    forall f x a, freshP mkFunPset a f -> freshP XP a x -> freshP YP a (f x).
  Proof.
    intros f x a [F1 [H J]] [F2 [K L]].
    exists (union F1 F2); split.
    auto with sets.
    apply supports_application; trivial.
  Qed.

End FunctionPermutations.

Implicit Arguments mkFunPset [A X Y].

Notation "X ^-> Y" := (mkFunPset X Y) (at level 90, right associativity).

Hint Resolve perm_fun perm_app : nominal.
Hint Resolve supports_fun : nominal.
Hint Resolve supports_application freshP_application : nominal.


Section FunctionPermutationFacts.
  Variable A : AtomT.
  Variables X Y : Set.
  Variables (XP : PsetT A X) (YP : PsetT A Y).

  Theorem function_constancy_v1 :
    forall (F : aset A) (f : A -> X),
    supports (mkAtomPset A ^-> XP) F f ->
      forall (a : A), ~ In a F -> freshP XP a (f a) ->
      forall (b : A), ~ In b F -> f a = f b.
  Proof.
    intros F f H a' M1 M2 b Q.

    case (atom_eqdec A b a'); intro L; try congruence.
    assert (freshP (mkAtomPset A) b a'); auto with nominal.
    assert (freshP (mkAtomPset A ^-> XP) b f); eauto with nominal.

    assert (freshP XP b (f a')).
    apply freshP_application with (XP := mkAtomPset A); trivial. 

    assert (Q1 : f a' = perm XP [(b, a')] (f a')).
    rewrite swap_freshP_atoms; trivial.

    rewrite perm_app with (XP := mkAtomPset A) in Q1.
    rewrite swap_non_supports with (F := F) in Q1; trivial.
    simpl in Q1; autorewrite with nominal in Q1; congruence.
  Qed.

  Theorem function_constancy_v2 :
    forall (F : aset A) (f : A -> X),
    supports (mkAtomPset A ^-> XP) F f ->
    (exists a : A, ~ In a F /\ freshP XP a (f a)) ->
    forall a b : A, ~ In a F -> ~ In b F -> f a = f b.
  Proof.
    intros F f H J c d C D.
    inversion_clear J as [a' [M1 M2]].
    rewrite <- (function_constancy_v1 F f H a' M1 M2 c); trivial.
    rewrite <- (function_constancy_v1 F f H a' M1 M2 d); trivial.
  Qed.

End FunctionPermutationFacts.

Hint Immediate function_constancy_v1 function_constancy_v2 : nominal.
