Require Import Arith.
Require Import Axioms.
Require Import AuxLib.
Require Import Nominal.
Require Import STLC.
Require Import Wf_nat.


Module MakeSTLC <: STLC.


(* *********************************************************************** *)
(** ** Basic definitions *)

(**
  Term variables.
*)
Definition tmvar := mkAtomRec 0.
Definition tmvarP := mkAtomPset tmvar.


(**
  STLC types.
*)
Inductive ty : Set :=
  | unit : ty
  | arrow : ty -> ty -> ty.

Definition tyP := mkIdPset tmvar ty.


(**
  [Phi n] is the set of all terms whose bound variables have index
  less than [n].  Thus, [Phi 0] corresponds to those terms with no
  dangling bound variables.  Free variables are still represented
  using names.
*)
Inductive Phi : nat -> Set :=
  | pdot : forall (n : nat), Phi n
  | pfree : forall (n : nat), tmvar -> Phi n
  | pbound : forall (n i : nat), i < n -> Phi n
  | papp : forall (n : nat), Phi n -> Phi n -> Phi n
  | plam : forall (n : nat), ty -> Phi (S n) -> Phi n.


(**
  We need to take advantage of proof irrelevance when considering
  equality on [Phi].
*)
Lemma pbound_eq :
  forall (n i i': nat), (i = i') ->
  forall (p : i < n) (q : i' < n),
  pbound n i p = pbound n i' q.
Proof.
  intros n i i' eq; dependent rewrite eq.
  intros p q; rewrite (Proof_irrelevance _ p q); reflexivity.
Qed.

Hint Resolve pbound_eq : core nominal.


(**
  Free variables.
*)
Fixpoint Phi_fv (n : nat) (x : Phi n) {struct x} : aset tmvar :=
  match x with
    | pdot _ => empty
    | pfree _ a => singleton a
    | pbound _ _ _ => empty
    | papp _ s t => union (Phi_fv _ s) (Phi_fv _ t)
    | plam _ _ s => Phi_fv _ s
  end.

Implicit Arguments Phi_fv [n].


(**
  Size.  The fact that [1 + x] converts to [S x] is helpful when defining
  the recursion combinator below, since it proceeds by induction on natural
  numbers (the initial value being the size of a term).
*)
Fixpoint Phi_size (n : nat) (x : Phi n) {struct x} : nat :=
  match x with
    | pdot _ => 1
    | pfree _ _ => 1
    | pbound _ _ _ => 1
    | papp _ s t => 1 + (Phi_size _ s) + (Phi_size _ t)
    | plam _ _ s => 1 + (Phi_size _ s)
  end.

Implicit Arguments Phi_size [n].


Lemma Phi_size_eq_O :
  forall (T : Type) (n : nat) (x : Phi n), Phi_size x = 0 -> T.
Proof.
  induction x; intros; discriminate.
Qed.


Lemma Phi_size_lt_O :
  forall (T : Type) (n : nat) (x : Phi n), Phi_size x <= 0 -> T.
Proof.
  intros T n x H.
  assert (J : Phi_size x = 0); auto with arith.
  apply Phi_size_eq_O with (x := x); trivial.
Qed.


(**
  We need ``weakening'' on terms in order to define substitution:
  Any term in [Phi n] is also a term in [Phi (S n)] modulo the proofs
  carried by [pbound] subterms.
*)
Fixpoint Phi_weaken (n : nat) (x : Phi n) {struct x} : Phi (S n) :=
  match x in Phi n return Phi (S n) with
    | pdot _ => pdot _
    | pfree _ a => pfree _ a
    | pbound _ i pf => pbound _ i (le_S _ _ pf)
    | papp _ s t => papp _ (Phi_weaken _ s) (Phi_weaken _ t)
    | plam _ T s => plam _ T (Phi_weaken _ s)
  end.

Implicit Arguments Phi_weaken [n].


(**
  Case analysis.  This lemma is useful when built-in tactics fail to provide
  similar functionality, for example when you want to analyze [x : Phi (S n)]
  and [case] fails.
*)
Lemma Phi_case :
  forall (n : nat) (P : Phi n -> Type),
  (P (pdot n)) ->
  (forall a, P (pfree n a)) ->
  (forall i pf, P (pbound n i pf)) ->
  (forall s t, P (papp n s t)) ->
  (forall T s, P (plam n T s)) ->
  (forall x, P x).
Proof.
  induction x; trivial.
Defined.


(**
  Well-founded induction on size.
*)
Lemma Phi_ind_size_multi :
  forall (P : forall (n : nat), Phi (S n) -> Type),
  (forall n x, (forall m y, Phi_size y < Phi_size x -> P m y) -> P n x) ->
  (forall n x, P n x).
Proof.
  intros P IH n x.
  assert (forall k n x, Phi_size x = k -> P n x); eauto.
    intros m; pattern m; apply (well_founded_induction_type lt_wf).
    intros k IHwf n' y eq; subst k; apply IH; intros p z LT.
    apply (IHwf (Phi_size z)); trivial.
Qed.


Lemma Phi_ind_size_single :
  forall (n : nat) (P : Phi n -> Type),
  (forall x, (forall y, Phi_size y < Phi_size x -> P y) -> P x) ->
  (forall x, P x).
Proof.
  intros n P IH x.
  assert (forall m x, Phi_size x = m -> P x); eauto.
    intros m; pattern m; apply (well_founded_induction_type lt_wf).
    intros k IHwf y eq; subst k; apply IH; intros z LT.
    apply (IHwf (Phi_size z)); trivial.
Qed.


(* *********************************************************************** *)
(** ** Permutations *)

Fixpoint Phi_perm
  (n : nat) (p : permt tmvar) (x : Phi n) {struct x} : Phi n
:=
  match x in Phi n return Phi n with
    | pdot _ => pdot _
    | pfree _ a => pfree _ (perm tmvarP p a)
    | pbound _ i pf => pbound _ i pf
    | papp _ s t => papp _ (Phi_perm _ p s) (Phi_perm _ p t)
    | plam _ T s => plam _ T (Phi_perm _ p s)
  end.

Implicit Arguments Phi_perm [n].


Definition PhiPermR (n : nat) : PsetT tmvar (Phi n).
  intros n; refine (mkPset (@Phi_perm n) _ _).
  induction x; simpl; autorewrite with nominal; congruence.
  intros p q; induction x; simpl; autorewrite with nominal; congruence.
Defined.


(**
  Permutation preserves the size of a term.
*)
Lemma Phi_size_perm :
  forall (n : nat) (p : permt tmvar) (x : Phi n),
  Phi_size (Phi_perm p x) = Phi_size x.
Proof.
  intros n p.
  induction x as [n | n a | n i pf | n s IHs t IHt | n T s IHs ]; simpl;
    congruence.
Qed.


(* *********************************************************************** *)
(** ** Abstraction and instantiation *)

(**
  Turn a free variable into a bound variable (an index).  We turn a term
  in [Phi n] into a term in [Phi (S n)] and choose [n] as the new index.
*)
Fixpoint abs
  (n : nat) (a : tmvar) (x : Phi n) {struct x}
  :  Phi (S n)
:=
  match x in Phi n return Phi (S n) with
    | pdot _ => pdot _
    | pfree i a' =>
        if atom_eqdec tmvar a a' then pbound (S i) i (lt_n_Sn _)
        else pfree _ a'
    | pbound i m pf => pbound (S i) m (lt_S _ _ pf)
    | papp _ s t => papp _ (abs _ a s) (abs _ a t)
    | plam _ T s => plam _ T (abs _ a s)
  end.

Implicit Arguments abs [n].


(**
  Turn the greatest index (bound variable) in a term into a free variable.
  A more general version of this function would instantiate the index with
  an arbitrary term.
*)
Definition inst_aux
  (n' : nat) (a : tmvar) (x : Phi n') (n : nat) (wf : n' = S n) : Phi n.
Proof.
  intros n' a x.
  induction x as [i | i a' | i m pf | i s IHs t IHt | i T s IHs].

  (* Case: pdot i *)
  intros; exact (pdot n).

  (* Case: pfree i a' *)
  intros; exact (pfree n a').

  (* Case: pbound i m pf *)
  intros n wf; rewrite wf in pf.
  case (eq_nat_dec m n).

      (* Subcase: m = n *)
      intro; exact (pfree n a).

      (* Subcase: m <> n *)
      intro H; exact (pbound n m (lt_Sn_n m n pf H)).

  (* Case: papp i s t *)
  intros n H; exact (papp n (IHs n H) (IHt n H)).

  (* Case: plam i s *)
  intros n H; exact (plam n T (IHs (S n) (eq_S _ _ H))).
Defined.


Definition inst (n : nat) (a : tmvar) (x : Phi (S n)) : Phi n :=
  inst_aux (S n) a x n (refl_equal _).

Implicit Arguments inst [n].


(**
  The following rewrites make it easier to work with [inst].  It is
  not usually helpful to unfold [inst] to [inst_aux].
*)
Lemma inst_papp :
  forall (n : nat) (a : tmvar) (s t : Phi (S n)),
  inst a (papp _ s t) = papp _ (inst a s) (inst a t).
Proof.
  reflexivity.
Qed.


Lemma inst_plam :
  forall (n : nat) (a : tmvar) (s : Phi (S (S n))) (T : ty),
  inst a (plam _ T s) = plam _ T (inst a s).
Proof.
  reflexivity.
Qed.


(**
  [abs] preserves the size of terms.
*)
Lemma Phi_size_abs :
  forall (n : nat) (x : Phi n) (a : tmvar),
  Phi_size (abs a x) = Phi_size x.
Proof.
  intros n.
  induction x as [n | n a' | n i pf | n s IHs t IHt | n T s IHs ]; simpl.
  reflexivity.
  intros a; case (atom_eqdec tmvar a a'); simpl; reflexivity.
  reflexivity.
  intros a; rewrite IHs; rewrite IHt; reflexivity.
  intros a; rewrite IHs; reflexivity.
Qed.


(**
  [inst] preserves the size of terms.
*)
Lemma Phi_size_inst :
  forall (n : nat) (x : Phi (S n)) (a : tmvar),
  Phi_size (inst a x) = Phi_size x.
Proof.
  intros n x.
  apply Phi_ind_size_multi
    with (P := fun (n : nat) (x : Phi (S n)) =>
                  forall a, Phi_size (inst a x) = Phi_size x); clear n x.
  intros n x; pattern x; apply Phi_case; clear x.
  reflexivity.
  reflexivity.
  intros i pf IH a; unfold inst; simpl; case (eq_nat_dec i n); reflexivity.
  intros s t IH a; rewrite inst_papp; simpl.
    rewrite (IH _ s); [idtac | simpl; auto with arith].
    rewrite (IH _ t); [idtac | simpl; auto with arith].
    reflexivity.
  intros T s IH a; rewrite inst_plam; simpl.
    rewrite (IH _ s); [idtac | simpl; auto with arith].
    reflexivity.
Qed.


(**
  Permutations commute with [abs].
*)
Lemma Phi_perm_abs :
  forall n p c (s : Phi n),
  Phi_perm p (abs c s) = abs (perm tmvarP p c) (Phi_perm p s).
Proof.
  intros n p c s.
  induction s as [n | n a | n i pf | n s IHs t IHt | n T s IHs]; simpl.
  reflexivity.
  atom_eqdec_case H c a;
    atom_eqdec_case J (p @ c) (p @ a); try reflexivity.
    assert (c = a); [eauto with nominal | congruence].
  reflexivity.
  rewrite IHs; rewrite IHt; reflexivity.
  rewrite IHs; reflexivity.
Qed.


(**
  Swapping pseudo-commutes with [inst] under certain conditions.
*)
Lemma Phi_swap_inst :
  forall (n : nat) (x : Phi (S n)) (a c d : tmvar),
  ~ In c (Phi_fv x) -> ~ In d (Phi_fv x) ->
  perm (PhiPermR n) [(c, d)] (inst a x) = inst (perm tmvarP [(c, d)] a) x.
Proof.
  intros n x.
  apply Phi_ind_size_multi
    with (P := fun (n : nat) (x : Phi (S n)) =>
            forall (a c d : tmvar),
            ~ In c (Phi_fv x) ->
            ~ In d (Phi_fv x) ->
            perm (PhiPermR n) [(c, d)] (inst a x) =
            inst (perm tmvarP [(c, d)] a) x); clear n x.
  intros n x; pattern x; apply Phi_case; clear x.
  intros IH a c d C D; unfold inst; simpl; reflexivity.
  intros a' IH a c d C D; unfold inst; simpl.
    autorewrite with nominal.
    swapa_case H J c d a'.
      simpl in C; assert (c <> a'); [auto with sets | intuition].
      simpl in D; assert (d <> a'); [auto with sets | intuition].
      reflexivity.
  intros i pf IH a c d C D; unfold inst; simpl;
    case (eq_nat_dec i n); reflexivity.
  intros s t IH a c d C D; rewrite inst_papp; simpl in * |- *.
    rewrite (IH _ s); auto with sets arith.
    rewrite (IH _ t); auto with sets arith.
  intros T s IH a c d C D; rewrite inst_plam; simpl in * |- *.
    rewrite (IH _ s); auto with sets.
Qed.


(**
  Swapping commutes with [inst] under certain conditions.
*)
Lemma Phi_swap_inst_commute :
  forall (n : nat) (x : Phi (S n)) (a c d : tmvar),
  a <> c -> a <> d ->
  perm (PhiPermR n) [(c, d)] (inst a x) =
  inst a (perm (PhiPermR (S n)) [(c, d)] x).
Proof.
  intros n x.
  apply Phi_ind_size_multi
    with (P := fun (n : nat) (x : Phi (S n)) =>
            forall (a c d : tmvar), a <> c -> a <> d ->
            perm (PhiPermR n) [(c, d)] (inst a x) =
            inst a (perm (PhiPermR (S n)) [(c, d)] x)); clear n x.
  intros n x; pattern x; apply Phi_case; clear x.
  intros IH a c d C D; unfold inst; simpl; reflexivity.
  intros a' IH a c d C D; unfold inst; simpl; reflexivity.
  intros i pf IH a c d C D; unfold inst; simpl.
    case (eq_nat_dec i n); intros EQ.
    simpl; autorewrite with nominal; reflexivity.
    simpl; auto.
  intros s t IH a c d C D; rewrite inst_papp; simpl in * |- *.
    rewrite (IH _ s); auto with sets arith.
    rewrite (IH _ t); auto with sets arith.
  intros T s IH a c d C D; rewrite inst_plam; simpl in * |- *.
    rewrite (IH _ s); auto with sets.
Qed.


(**
  [inst] is the inverse of [abs].
*)
Lemma inst_abs_inv :
  forall (n : nat) (x : Phi n) (a : tmvar), inst a (abs a x) = x.
Proof.
  intros n; induction x as [n | n a | n i pf | n s IHs t IHt | n T s IHs].
  intros a'; reflexivity.
  intros a'; simpl.
    atom_eqdec_case H a' a; unfold inst; simpl; try reflexivity.
    case (eq_nat_dec n n); intros; congruence.
  unfold inst; simpl; case (eq_nat_dec i n); intuition.
    assert False; [omega | intuition].
  intros a'; simpl; rewrite inst_papp; rewrite IHs; rewrite IHt; reflexivity.
  intros a'; simpl; rewrite inst_plam; rewrite IHs; reflexivity.
Qed.


(**
  [abs] is the inverse of [inst] under certain conditions.
*)
Lemma abs_inst_inv :
  forall (n : nat) (x : Phi (S n)) (a : tmvar),
  ~ In a (Phi_fv x) -> abs a (inst a x) = x.
Proof.
  intros n x.
  apply Phi_ind_size_multi
    with (P := fun (n : nat) (x : Phi (S n)) =>
            forall (a : tmvar),
            ~ In a (Phi_fv x) -> abs a (inst a x) = x); clear n x.
  intros n x; pattern x; apply Phi_case; clear x.
  reflexivity.
  intros a' IH a H; simpl in H |- *.
    atom_eqdec_case J a a'.
      assert (a <> a'); [auto with sets | intuition].
      reflexivity.
  intros i pf IH a H; unfold inst; simpl in H |- *.
    case (eq_nat_dec i n); simpl; auto with nominal.
      intros eq''; atom_eqdec_case H a a; auto.
  intros s t IH a H; rewrite inst_papp; simpl in H |- *.
    rewrite (IH _ s); simpl; auto with arith sets.
    rewrite (IH _ t); simpl; auto with arith sets.
  intros T s IH a H; rewrite inst_plam; simpl in H |- *.
    rewrite (IH _ s); simpl; auto with sets.
Qed.


(**
  The free variables of [abs a x] are the obvious ones.
*)
Lemma Phi_fv_abs :
  forall (n : nat) (x : Phi n) (a : tmvar),
  Phi_fv (abs a x) = remove a (Phi_fv x).
Proof.
  intros n x; induction x as [n | n a' | n i pf | n s IHs t IHt | n T s IHs];
    intro a; simpl.
  autorewrite with sets; auto.
  case (atom_eqdec tmvar a a'); intro H; simpl.
    autorewrite with sets; auto.
    autorewrite with sets; auto.
  autorewrite with sets; auto.
  rewrite IHs; rewrite IHt; autorewrite with sets; auto.
  rewrite IHs; autorewrite with sets; auto.
Qed.


(* *********************************************************************** *)
(** ** User-level abstraction for terms *)

Definition tm : Set := Phi 0.
Definition tmP := PhiPermR 0.
Definition dot : tm := pdot 0.
Definition var (a : tmvar) : tm := pfree 0 a.
Definition app (s t : tm) : tm:= papp 0 s t.
Definition lam (x : tmvar) (T : ty) (b : tm) : tm := plam 0 T (abs x b).


(* *********************************************************************** *)
(** ** Permutations *)

Section STLCPermutations.
Variable p : permt tmvar.
Variable c : tmvar.
Variable T : ty.
Variables s t : tm.

Theorem perm_dot : perm tmP p dot = dot.
Proof.
  reflexivity.
Qed.

Theorem perm_var : perm tmP p (var c) = var (perm tmvarP p c).
Proof.
  reflexivity.
Qed.

Theorem perm_app :
  perm tmP p (app s t) = app (perm tmP p s) (perm tmP p t).
Proof.
  reflexivity.
Qed.

Theorem perm_lam :
  perm tmP p (lam c T s) = lam (perm tmvarP p c) T (perm tmP p s).
Proof.
  unfold lam; simpl.
  rewrite Phi_perm_abs; reflexivity.
Qed.

End STLCPermutations.


(* *********************************************************************** *)
(** ** Free variables *)

Definition fvar : tm -> aset tmvar := Phi_fv (n := 0).


Theorem fvar_dot : fvar dot = empty.
Proof.
  reflexivity.
Qed.


Theorem fvar_var : forall a, fvar (var a) = singleton a.
Proof.
  reflexivity.
Qed.


Theorem fvar_app : forall s t, fvar (app s t) = union (fvar s) (fvar t).
Proof.
  reflexivity.
Qed.


Theorem fvar_lam : forall a T s, fvar (lam a T s) = remove a (fvar s).
Proof.
  intros a T s; unfold fvar; unfold lam; simpl.
  apply Phi_fv_abs.
Qed.


(* *********************************************************************** *)
(** ** Alpha-equality *)

Lemma swap_in_abs :
  forall (n : nat) (s : Phi n) (a b : tmvar),
  b <> a ->
  ~ In a (Phi_fv (Phi_perm [(a, b)] s)) ->
  abs b (Phi_perm [(a, b)] s) = abs a s.
Proof.
  intros n s; induction s as [n | n a' | n i pf | n s IHs t IHt | n T s IHs];
    intros a b H; simpl.
  reflexivity.
  autorewrite with nominal.
    swapa_case Q R a b a'; atom_eqdec_simpl; try reflexivity.
      intros J; destruct_neg_singleton J; congruence.
  reflexivity.
  intros J; destruct_neg_union J Q R.
    rewrite IHs; trivial.
    rewrite IHt; trivial.
  intros J; rewrite IHs; trivial.
Qed.


Lemma eq_lam_diff :
  forall a b T T' s s',
  T = T' ->
  a <> b ->
   ~ In b (fvar s) ->
  s = perm tmP [(a, b)] s' ->
  lam a T s = lam b T' s'.
Proof.
  intros a b T T' s s' H J K L; rewrite H; unfold lam.
  rewrite L; rewrite perm_swap_switch; simpl.
  rewrite swap_in_abs; trivial.
  change (Phi_perm [(b, a)] s') with (perm tmP [(b, a)] s').
  rewrite perm_swap_switch; rewrite <- L; trivial.
Qed.


Theorem eq_lam :
  forall a b T s,
   ~ In b (fvar s) ->
  lam a T s = lam b T (perm tmP [(a, b)] s).
Proof.
  intros a b T s H.
  case (atom_eqdec tmvar a b); intros J.

    (* Case: a = b *)
    rewrite J; autorewrite with nominal.
    trivial.

    (* Case: a <> b *)
    apply eq_lam_diff; trivial.
    autorewrite with nominal; trivial.
Qed.


(* *********************************************************************** *)
(** ** Structural induction *)

Theorem tm_induction_weak :
  forall (P : tm -> Type),
  (P dot) ->
  (forall a, P (var a)) ->
  (forall s, P s -> forall t, P t -> P (app s t)) ->
  (forall a T s, P s -> P (lam a T s)) ->
  forall x : tm, P x.
Proof.
  intros P f_dot f_var f_app f_lam x.
  apply (Phi_ind_size_single 0); clear x.
  intros x; pattern x; apply Phi_case; clear x.
  intros IH; exact f_dot.
  intros a IH; exact (f_var a).
  intros i pf; assert False; [inversion pf | intuition].
  intros s t IH; fold (app s t); apply f_app.
    apply IH; simpl; auto with arith.
    apply IH; simpl; auto with arith.
  intros T s IH.
    case (atom_infinite tmvar (Phi_fv s)); intros a A.
    replace s with (abs a (inst a s)); fold (lam a T (inst a s)).
      apply f_lam.
        apply IH; rewrite Phi_size_inst; simpl; auto.
    apply abs_inst_inv; auto with sets.
Qed.


Theorem tm_induction :
  forall (P : tm -> Type) (F : aset tmvar),
  (P dot) ->
  (forall a, P (var a)) ->
  (forall s, P s -> forall t, P t -> P (app s t)) ->
  (forall a, ~ In a F -> forall T, forall s, P s -> P (lam a T s)) ->
  forall x : tm, P x.
Proof.
  intros P F f_dot f_var f_app f_lam x.
  apply (Phi_ind_size_single 0); clear x.
  intros x; pattern x; apply Phi_case; clear x.
  intros IH; exact f_dot.
  intros a IH; exact (f_var a).
  intros i pf; assert False; [inversion pf | intuition].
  intros s t IH; fold (app s t); apply f_app.
    apply IH; simpl; auto with arith.
    apply IH; simpl; auto with arith.
  intros T s IH.
    case (atom_infinite tmvar (union F (Phi_fv s))); intros a A.
    replace s with (abs a (inst a s)); fold (lam a T (inst a s)).
      apply f_lam; auto with sets.
        apply IH; rewrite Phi_size_inst; simpl; auto.
    apply abs_inst_inv; auto with sets.
Qed.


(* *********************************************************************** *)
(** ** Recursion *)

Section Recursion.

Variables (R : Set) (RP : PsetT tmvar R).
Variable (f_dot : R).
Variable (f_var : tmvar -> R).
Variable (f_app : tm -> R -> tm -> R -> R).
Variable (f_lam : tmvar -> ty -> tm -> R -> R).
Variable (F : aset tmvar).
Variable (supp_dot : supports RP F f_dot).
Variable (supp_var : supports (tmvarP ^-> RP) F f_var).
Variable (supp_app : supports (tmP ^-> RP ^-> tmP ^-> RP ^-> RP) F f_app).
Variable (supp_lam : supports (tmvarP ^-> tyP ^-> tmP ^-> RP ^-> RP) F f_lam).
Variable (fcb :
  exists b, ~ In b F /\ forall x y z, freshP RP b (f_lam b x y z)).


Derive Dependent Inversion tm_inv with (Phi 0) Sort Set.


Lemma Phi_size_lt_papp_fun :
  forall n m s t, Phi_size (papp n s t) <= S m -> Phi_size s <= m.
Proof.
  intros; simpl in * |- *; omega.
Qed.

Implicit Arguments Phi_size_lt_papp_fun [n m s t].


Lemma Phi_size_lt_papp_arg :
  forall n m s t, Phi_size (papp n s t) <= S m -> Phi_size t <= m.
Proof.
  intros; simpl in * |- *; omega.
Qed.

Implicit Arguments Phi_size_lt_papp_arg [n m s t].


Lemma Phi_size_lt_plam :
  forall n m a T s, Phi_size (plam n T s) <= S m -> Phi_size (inst a s) <= m.
Proof.
  intros; simpl in *; rewrite Phi_size_inst; auto with arith.
Qed.

Implicit Arguments Phi_size_lt_plam [n m a T s].


(**
  Our goal here is to define recursive functions over terms by induction
  on their size.  This translates to induction on a natural number which
  bounds the size of the term.  All the arguments here are important: [n]
  is required since the function is defined by induction on the structure
  of [n], [x] is the term we actually care about, and the proof ensures
  that we don't have to deal with non-sensical combinations of [n] and [x].

  The proof is of an [le] relation and not an [eq] relation due to how
  [recF] makes recursive calls.  We could, in theory, use an [eq] relation
  if we were willing to define [recF] by well-founded induction on [n], but
  that seems to pose its own problems.
*)
Definition recF :
  forall (n : nat) (x : Phi 0), Phi_size x <= n -> R.
Proof.
  intros n; induction n as [ | n IH ].

  (* Case : n = 0. *)
  intros x H.
  exact (Phi_size_lt_O _ _ _ H).

  (* Case: n = S n *)
  intros x; inversion x using tm_inv.

    (* Subcase: pdot *)
    intros _ m _ _; exact f_dot.

    (* Subcase: pfree *)
    intros _ m a _ _; exact (f_var a).

    (* Subcase: pbound *)
    intros _ m i _ pf.
    assert False; [inversion pf | intuition].

    (* Subcase: papp *)
    intros _ m _ s t LE.
    refine (f_app s (IH s _) t (IH t _)).
      exact (Phi_size_lt_papp_fun LE).
      exact (Phi_size_lt_papp_arg LE).

    (* Subcase: plam *)
    intros _ m T _ s LE.
    case (atom_infinite tmvar (union F (Phi_fv s))); intros a A.
    refine (f_lam a T (inst a s) (IH (inst a s) _)).
      exact (Phi_size_lt_plam LE).
Defined.


(**
  The particular [n] and proof supplied to [recF] don't matter.
*)
Lemma recF_stable :
  forall (n m : nat) (x : tm) (pfn : Phi_size x <= n) (pfm : Phi_size x <= m),
  recF n x pfn = recF m x pfm.
Proof.
  induction n as [ | n IHn]; induction m as [ | m IHm].
  intros x pfn; exact (Phi_size_lt_O _ _ _ pfn).
  intros x pfn; exact (Phi_size_lt_O _ _ _ pfn).
  intros x pfn pfm; exact (Phi_size_lt_O _ _ _ pfm).
  intros x; pattern x; apply (Phi_case 0).
  reflexivity.
  reflexivity.
  reflexivity.
  intros s t pfn pfm; simpl.
    apply f_equal4 with (f := f_app); trivial.
  intros T s pfn pfm; simpl.
    case (atom_infinite tmvar (union F (Phi_fv s))); intros a A.
    apply f_equal4 with (f := f_lam); trivial.
Qed.


Lemma recF_stable_gen :
  forall (n m : nat) (x y : tm), x = y ->
  forall (pfn : Phi_size x <= n) (pfm : Phi_size y <= m),
  recF n x pfn = recF m y pfm.
Proof.
  intros n m x y eq; dependent rewrite eq.
  intros pfn pfm; apply recF_stable.
Qed.

Hint Resolve recF_stable_gen : core.


(**
  Define the recursion combinator, as seen by the user.  Most
  of the cases for how it evaluates are straightforward.
*)
Definition tm_rec (x : tm) : R := recF (Phi_size x) x (le_n _).


Theorem tm_rec_dot : tm_rec dot = f_dot.
Proof.
  reflexivity.
Qed.


Theorem tm_rec_var : forall a, tm_rec (var a) = f_var a.
Proof.
  reflexivity.
Qed.


Theorem tm_rec_app :
  forall s t, tm_rec (app s t) = f_app s (tm_rec s) t (tm_rec t).
Proof.
  intros s t; unfold tm_rec; simpl.
  f_equal; auto.
Qed.


(**
  Proving the rewrite rule for the [lam] case requires that we prove
  the following two properties simultaneously.  There may be an easier
  way, but at least it doesn't seem that this particular method is
  specific to the object-language at hand.

  It's only at this point where we need to know the support of [f_dot],
  [f_var], [f_app], and [f_lam], and know that [fcb] holds.
*)
Let rewrite_prop (n : nat) :=
  forall a T s, Phi_size (lam a T s) <= n ->
  ~ In a F -> tm_rec (lam a T s) = f_lam a T s (tm_rec s).

Let swap_prop (n : nat) :=
  forall x c d pf pf', ~ In c F -> ~ In d F ->
  perm RP [(c, d)] (recF n x pf) = recF n (perm tmP [(c, d)] x) pf'.


Lemma rewrite_to_swap :
  forall (n : nat), (forall m, m <= n -> rewrite_prop m) -> swap_prop n.
Proof.
  (* XXX This proof is simply ugly... *)
  unfold swap_prop.
  induction n as [ | n IHn ]; intros RH.

  intros x c d pf.
  exact (Phi_size_lt_O _ _ _ pf).

  intros x; pattern x; apply tm_induction with (F := F).

  intros c d pf pf' C D; simpl.
  rewrite swap_non_supports with (F := F); trivial.

  intros a c d pf pf' C D; simpl.
  rewrite Psets.perm_app with (XP := tmvarP).
  rewrite swap_non_supports with (F := F); trivial.

  intros s _ t _ c d pf pf' C D; simpl.
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite swap_non_supports with (F := F); trivial.
  apply f_equal4 with (f := f_app); auto.
    assert (pf'' : Phi_size (perm tmP [(c, d)] s) <= n).
      simpl in pf' |- *; omega.
    rewrite IHn with (pf' := pf''); auto.
    assert (pf'' : Phi_size (perm tmP [(c, d)] t) <= n).
      simpl in pf' |- *; omega.
    rewrite IHn with (pf' := pf''); auto.

  intros a A T s _ c d pf pf' C D.
  unfold rewrite_prop, tm_rec in RH.
  match goal with |- ?X = ?Y => replace Y with
    (recF (Phi_size (lam (perm tmvarP [(c, d)] a) T (perm tmP [(c, d)] s)))
          (lam (perm tmvarP [(c, d)] a) T (perm tmP [(c, d)] s)) (le_n _)) end; auto.
  rewrite (RH (Phi_size (lam (perm tmvarP [(c, d)] a) T (perm tmP [(c, d)] s)))); auto.
  replace (recF (S n) (lam a T s) pf)
     with (recF (Phi_size (lam a T s)) (lam a T s) (le_n _)); auto.
  rewrite (RH (Phi_size (lam a T s))); auto.
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite Psets.perm_app with (XP := tyP).
  rewrite Psets.perm_app with (XP := tmvarP).
  rewrite swap_non_supports with (F := F) (x := f_lam); auto.
  apply f_equal4 with (f := f_lam); auto.
  assert (pfQ : Phi_size s <= n).
    simpl in pf; rewrite Phi_size_abs in pf; auto with arith.
  assert (pfQ' : Phi_size (perm tmP [(c, d)] s) <= n).
    simpl; rewrite Phi_size_perm; trivial.
  replace (recF (Phi_size s) s (le_n (Phi_size s)))
     with (recF n s pfQ); auto.
  rewrite IHn with (pf := pfQ) (pf' := pfQ'); auto.

  assert (forall m : nat, m <= S n -> rewrite_prop m); auto.
    unfold rewrite_prop, tm_rec; auto.
  rewrite <- perm_lam; simpl in pf |- *; trivial.
  unfold tmvarP; rewrite perm_atom; apply swap_not_in_set; trivial.
  apply recF_stable_gen; rewrite perm_lam; trivial.
Qed.


Lemma swap_to_rewrite :
  forall (n : nat), (forall m, m < n -> swap_prop m) -> rewrite_prop n.
Proof.
  case fcb; intros b [B Bfresh].
  induction n as [ | n IHn ].

  unfold rewrite_prop; intros H a T s J.
  exact (Phi_size_lt_O _ _ _ J).

  unfold rewrite_prop; intros H a T s J A.
  unfold tm_rec; simpl.
  case (atom_infinite tmvar (union F (Phi_fv (abs a s)))); intros a' A'.

  pose (g := fun a' => f_lam a' T (inst a' (abs a s))
                (recF (Phi_size (inst a' (abs a s)))
                      (inst a' (abs a s))
                      (le_n _))).
  pose (G := union F (Phi_fv (abs a s))).

  assert (Q : forall x y, ~ In x G -> ~ In y G -> g x = g y).
    intros x y X Y.
    apply function_constancy_v2 with (A := tmvar) (F := G) (XP := RP);
      trivial.

    apply supports_fun; unfold g, G in * |- *; simpl; intros u v z U V.
    rewrite Psets.perm_app with (XP := RP).
    rewrite Psets.perm_app with (XP := tmP).
    rewrite Psets.perm_app with (XP := tyP).
    rewrite Psets.perm_app with (XP := tmvarP).
    rewrite swap_non_supports with (F := F) (x := f_lam); auto with sets.
    apply f_equal4 with (f := f_lam); try trivial.
    unfold tmP; rewrite Phi_swap_inst; auto with sets.

    unfold swap_prop in H.
    assert (pfQ : Phi_size (perm tmP [(u, v)] (inst z (abs a s))) <=
                  Phi_size (inst z (abs a s))).
      simpl; rewrite Phi_size_perm; trivial.
    rewrite H with (pf' := pfQ); auto with sets.
    apply recF_stable_gen.
    unfold tmP; rewrite Phi_swap_inst; auto with sets.
    simpl in J; rewrite Phi_size_inst; trivial.

    case (atom_infinite tmvar G); intros a'' A''; exists a''.
    split; [trivial | unfold g].
    assert (Q : freshP RP (swapa tmvar (b, a'') a'')
               (perm RP [(b, a'')] (f_lam a'' T (inst a'' (abs a s))
      (recF (Phi_size (inst a'' (abs a s))) (inst a'' (abs a s)) (le_n _))))).

      unfold G in A''.
      swapa_simpl.
      rewrite Psets.perm_app with (XP := RP).
      rewrite Psets.perm_app with (XP := tmP).
      rewrite Psets.perm_app with (XP := tyP).
      rewrite Psets.perm_app with (XP := tmvarP).
      rewrite swap_non_supports with (F := F) (x := f_lam); auto with sets.
      simpl; repeat rewrite perm_atom; autorewrite with nominal; swapa_simpl; trivial.

    replace (swapa tmvar (b, a'') a'')
       with (perm tmvarP [(b, a'')] a'') in Q.
    eapply perm_freshP'; eexact Q.
    unfold tmvarP; rewrite perm_atom.
    autorewrite with nominal; reflexivity.

  match goal with |- ?X = ?Y => replace X with (g a') end.
  match goal with |- ?X = ?Y => replace Y with (g a) end.
    apply Q; trivial.
      unfold G; rewrite Phi_fv_abs; auto with sets.
    unfold g; rewrite inst_abs_inv; reflexivity.
    unfold g; apply f_equal4 with (f := f_lam); auto.
Qed.


Lemma rewrite_and_swap : forall m, rewrite_prop m /\ swap_prop m.
Proof.
  intros m; pattern m; apply (well_founded_ind lt_wf); clear m.
  intros n IH.
  assert (rewrite_prop n).
    apply swap_to_rewrite.
    intros; eapply proj2; eauto.
  split; trivial.
  apply rewrite_to_swap.
  intros x LE.
  case (le_lt_eq_dec _ _ LE).
    intros J; eapply proj1; eauto.
    intros J; rewrite J; trivial.
Qed.


Theorem tm_rec_lam :
  forall a T s, ~ In a F -> tm_rec (lam a T s) = f_lam a T s (tm_rec s).
Proof.
  intros a T s H.
  assert (Q := proj1 (rewrite_and_swap (Phi_size (lam a T s)))).
  unfold rewrite_prop in Q; apply Q; trivial.
Qed.


Theorem tm_rec_supp :
  supports (tmP ^-> RP) F tm_rec.
Proof.
  apply supports_fun; intros a b x A B.
  pattern x; apply tm_induction with (F := F); clear x.

  rewrite perm_dot; rewrite tm_rec_dot.
  rewrite swap_non_supports with (F := F); trivial.

  intros a'.
  rewrite perm_var; repeat rewrite tm_rec_var.
  rewrite Psets.perm_app with (XP := tmvarP).
  rewrite swap_non_supports with (F := F); trivial.

  intros s IHs t IHt.
  rewrite perm_app; repeat rewrite tm_rec_app.
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite swap_non_supports with (F := F); trivial.
  rewrite IHs; rewrite IHt; trivial.

  intros a' A' T s IHs.
  rewrite perm_lam; repeat rewrite tm_rec_lam; trivial.
  rewrite Psets.perm_app with (XP := RP).
  rewrite Psets.perm_app with (XP := tmP).
  rewrite Psets.perm_app with (XP := tyP).
  rewrite Psets.perm_app with (XP := tmvarP).
  rewrite swap_non_supports with (F := F); trivial.
  rewrite IHs; trivial.
  unfold tmvarP; auto with nominal.
Qed.

End Recursion.


(* *********************************************************************** *)
(** ** Constructor distinctness *)

Section Distinctness.
Variables a b : tmvar.
Variables T : ty.
Variables s t q r : tm.

Theorem distinctness_dot_var: dot <> var a.
Proof. discriminate. Qed.

Theorem distinctness_dot_app : dot <> app s t.
Proof. discriminate. Qed.

Theorem distinctness_dot_lam : dot <> lam a T s.
Proof. discriminate. Qed.

Theorem distinctness_var_dot : var a <> dot.
Proof. discriminate. Qed.

Theorem distinctness_var_app : var a <> app s t.
Proof. discriminate. Qed.

Theorem distinctness_var_lam : var a <> lam b T s.
Proof. discriminate. Qed.

Theorem distinctness_app_dot : app s t <> dot.
Proof. discriminate. Qed.

Theorem distinctness_app_var : app s t <> var a.
Proof. discriminate. Qed.

Theorem distinctness_app_lam : app s t <> lam a T q.
Proof. discriminate. Qed.

Theorem distinctness_lam_dot : lam a T s <> dot.
Proof. discriminate. Qed.

Theorem distinctness_lam_var : lam a T s <> var b.
Proof. discriminate. Qed.

Theorem distinctness_lam_app : lam a T s <> app q r.
Proof. discriminate. Qed.

End Distinctness.


(* *********************************************************************** *)
(** ** Injectivity for constructors *)

(**
  Some of the lemmas in this section are needed because the [injection]
  tactic fails to produce useful hypotheses.
*)
Theorem injection_var : forall a b, var a = var b -> a = b.
Proof.
  intros a b H; injection H; auto.
Qed.


Lemma injection_papp_fun :
  forall n s s' t t', papp n s t = papp n s' t' -> s = s'.
Proof.
  intros n s s' t t' H.
  set (f := fun x : Phi n =>
      (match x in Phi i return (i = n) -> Phi i with
        | papp i s t => fun _ => s
        | pdot i => fun pf => eq_rec_r (fun n => Phi n) t pf
        | pfree i _ => fun pf => eq_rec_r (fun n => Phi n) t pf
        | pbound i _ _ => fun pf => eq_rec_r (fun n => Phi n) t pf
        | plam i _ _ => fun pf => eq_rec_r (fun n => Phi n) t pf
      end) (refl_equal n)).
  change (f (papp n s t) = f (papp n s' t')).
  rewrite H; reflexivity.
Qed.


Theorem injection_app_fun : forall s s' t t', app s t = app s' t' -> s = s'.
Proof.
  unfold app; intros s s' t t' H.
  exact (injection_papp_fun 0 s s' t t' H).
Qed.


Lemma injection_papp_arg :
  forall n s s' t t', papp n s t = papp n s' t' -> t = t'.
Proof.
  intros n s s' t t' H.
  set (f := fun x : Phi n =>
      (match x in Phi i return (i = n) -> Phi i with
        | papp i s t => fun _ => t
        | pdot i => fun pf => eq_rec_r (fun n => Phi n) t pf
        | pfree i _ => fun pf => eq_rec_r (fun n => Phi n) t pf
        | pbound i _ _ => fun pf => eq_rec_r (fun n => Phi n) t pf
        | plam i _ _ => fun pf => eq_rec_r (fun n => Phi n) t pf
      end) (refl_equal n)).
  change (f (papp n s t) = f (papp n s' t')).
  rewrite H; reflexivity.
Qed.


Theorem injection_app_arg : forall s s' t t', app s t = app s' t' -> t = t'.
Proof.
  unfold app; intros s s' t t' H.
  exact (injection_papp_arg 0 s s' t t' H).
Qed.


Theorem injection_lam_ty :
  forall a a' T T' s s',
  lam a T s = lam a' T' s' ->
  (T = T').
Proof.
  intros a a' T T' s s' H.
  set (f := fun x : Phi 0 =>
        match x with
          | papp i s t => T
          | pdot i => T
          | pfree i _ => T
          | pbound i _ _ => T
          | plam i T _ => T
        end).
  change (f (lam a T s) = f (lam a' T' s')).
  rewrite H; reflexivity.
Qed.


Lemma injection_plam :
  forall (n : nat) (T T' : ty) (s s' : Phi (S n)),
  plam n T s = plam n T' s' -> s = s'.
Proof.
  intros n T T' s s' H.
  set (f := fun x : Phi n =>
      (match x in Phi i return (i = n) -> Phi (S i) with
        | papp i _ _ => fun pf => eq_rec_r _ s (eq_S _ _ pf)
        | pdot i => fun pf => eq_rec_r _ s (eq_S _ _ pf)
        | pfree i _ => fun pf => eq_rec_r _ s (eq_S _ _ pf)
        | pbound i _ _ => fun pf => eq_rec_r _ s (eq_S _ _ pf)
        | plam i _ s => fun _ => s
      end) (refl_equal n)).
  change (f (plam n T s) = f (plam n T' s')).
  rewrite H; reflexivity.
Qed.


Lemma abs_eq :
  forall (n : nat) (a : tmvar) (s s' : Phi n),
  abs a s = abs a s' -> s = s'.
Proof.
  (* XXX This proof is simply ugly... *)
  intros n a.
  induction s as [n | n a' | n i pf | n s IHs t IHt | n T s IHs];
    induction s' as
      [n' | n' a'' | n' i' pf' | n' s' IHs' t' IHt' | n' T' s' IHs'];
    simpl; auto;
    try (atom_eqdec_case Q a a');
    try (atom_eqdec_case R a a'');
    try (intro Z; inversion Z; fail).

  congruence.
  intros H; injection H; congruence.
  intros H; injection H; intros; assert False; [omega | intuition].
  intros H; injection H; intros; assert False; [omega | intuition].
  intros H; injection H; auto.

  intros eq.
  assert (abs a s = abs a s'); [eapply injection_papp_fun; eauto | idtac].
  assert (abs a t = abs a t'); [eapply injection_papp_arg; eauto | idtac].
  assert (Q : s = s'); eauto.
  assert (R : t = t'); eauto.
  congruence.

  intros eq; injection eq; intros _ Q.
  assert (abs a s = abs a s'); [eapply injection_plam; eauto | idtac].
  assert (R : s = s'); eauto.
  congruence.
Qed.


Lemma abs_neq :
  forall (n : nat) (a a' : tmvar) (s s' : Phi n),
  a <> a' ->
  abs a s = abs a' s' ->
  ~ In a (Phi_fv s') /\ s = Phi_perm [(a, a')] s'.
Proof.
  (* XXX This proof is simply ugly... *)
  intros n b b'.
  induction s as [n | n a' | n i pf | n s IHs t IHt | n T s IHs];
    induction s' as
      [n' | n' a'' | n' i' pf' | n' s' IHs' t' IHt' | n' T' s' IHs'];
    intro neq; simpl;
    try (atom_eqdec_case H b' a'');
    try (atom_eqdec_case J b a');
    try (intro Z; inversion Z; fail);
    try subst;
    try (intro Z; injection Z; intros; subst;
      autorewrite with nominal; swapa_simpl; auto with sets ; fail);
    autorewrite with nominal; swapa_simpl; auto with sets.

  intro Z; injection Z; intros; subst.
    assert False; [ omega | intuition ].
  intro Z; injection Z; intros; subst.
    assert False; [ omega | intuition ].

  intro L.
    assert (abs b s = abs b' s'); [eapply injection_papp_fun; eauto | idtac].
    assert (abs b t = abs b' t'); [eapply injection_papp_arg; eauto | idtac].
    assert (Q : ~ In b (Phi_fv s') /\ s = Phi_perm [(b, b')] s'); auto.
    assert (R : ~ In b (Phi_fv t') /\ t = Phi_perm [(b, b')] t'); auto.
    inversion_clear Q as [Q1 Q2]; inversion_clear R as [R1 R2].
    rewrite Q2; rewrite R2; auto with sets.

  intro L.
    injection L; intros _ M; rewrite M.
    assert (abs b s = abs b' s'); [eapply injection_plam; eauto | idtac].
    assert (Q : ~ In b (Phi_fv s') /\ s = Phi_perm [(b, b')] s'); auto.
    inversion_clear Q as [Q1 Q2].
    rewrite Q2; auto with sets.
Qed.


Theorem injection_lam_body :
  forall a a' T T' s s',
  lam a T s = lam a' T' s' ->
  (a = a' /\ s = s') \/
  (a <> a' /\ ~ In a (fvar s') /\ s = perm tmP [(a, a')] s').
Proof.
  intros a a' T T' s s' H.
  case (atom_eqdec tmvar a a'); intro J.

    (* a = a' *)
    left; split; auto.
    rewrite J in H; unfold lam in H; unfold tm in *.
    assert (Q : abs a' s = abs a' s').
      eapply injection_plam; eauto.
      eapply abs_eq; eauto.

    (* a <> a' *)
    right; unfold fvar.
    unfold lam in H.
    injection H; intros _ K; rewrite K in H; simpl.
    assert (Q : abs a s = abs a' s').
      eapply injection_plam; eauto.
      split; auto.
      eapply abs_neq with (s := s) (s' := s'); auto.
Qed.


End MakeSTLC.
