Require Import Axioms.
Require Export Swaps.


(* *********************************************************************** *)
(** ** Support *)

Section Support.

Variable AR : AtomT.
Variable X : Set.
Variable S : SwapT AR X.


Definition supports (E : aset AR) (x : X) : Prop :=
  forall (a b : atom AR), ~ In a E -> ~ In b E -> swap S (a, b) x = x.

Definition fresh (a : atom AR) (x : X) : Prop :=
  exists E : aset AR, supports E x /\ ~ In a E.


Lemma swap_non_supports :
  forall (F : aset AR) (a b : atom AR) (x : X),
  supports F x -> ~ In a F -> ~ In b F -> swap S (a, b) x = x.
Proof.
  unfold supports; intuition.
Qed.


Lemma supports_subset :
  forall (E F : aset AR) (x : X),
  supports E x -> Subset E F -> supports F x.
Proof.
  unfold supports; intros E F x H J a b A B.
  rewrite H; intuition.
Qed.


Lemma supports_intersection :
  forall (E F : aset AR) (x : X),
  supports E x -> supports F x -> supports (intersection E F) x.
Proof.
  unfold supports; intros E F x H J a a' A B.
  case (atom_eqdec AR a a'); intro T.

  (* Case: a = a' *)
  rewrite T; rewrite swap_same; auto.

  (* Case: a <> a' *)
  case (atom_infinite AR (add a' (add a (union E F)))); intros a'' M.
  assert (M1 : ~ In a'' E).
    eapply union_neg_elim_1; do 2 eapply add_neg_elim_2; eauto with sets.
  assert (M2 : ~ In a'' F).
    eapply union_neg_elim_2; do 2 eapply add_neg_elim_2; eauto with sets.
  assert (M3 : a'' <> a).
    apply add_neg_elim_1 with (E := union E F).
    apply add_neg_elim_2 with (y := a'); intuition.
  assert (M4 : a'' <> a').
    eapply add_neg_elim_1 with (E := add a (union E F)); intuition.

  assert (Q : swap S (a, a') x =
              swap S (a, a'') (swap S (a', a'') (swap S (a, a'') x))).
  rewrite swap_distrib; swapa_simpl; autorewrite with nominal.
  rewrite swap_switch; reflexivity.

  rewrite Q.
  assert (Q1 : ~ In a E \/ ~ In a F); auto with sets.
  assert (Q2 : ~ In a' E \/ ~ In a' F); auto with sets.
  inversion_clear Q1 as [U | U]; inversion_clear Q2 as [V | V].

  repeat rewrite H; trivial.

  replace (swap S (a, a'') x) with x; [idtac | rewrite H; trivial].
  replace (swap S (a', a'') x) with x; [idtac | rewrite J; trivial].
  replace (swap S (a, a'') x) with x; [auto | rewrite H; trivial].

  replace (swap S (a, a'') x) with x; [idtac | rewrite J; trivial].
  replace (swap S (a', a'') x) with x; [idtac | rewrite H; trivial].
  replace (swap S (a, a'') x) with x; [auto | rewrite J; trivial].

  repeat rewrite J; trivial.
Qed.

Hint Resolve supports_intersection : nominal.



Lemma supports_to_fresh :
  forall (E : aset AR) (a : atom AR) (x : X),
  supports E x -> ~ In a E -> fresh a x.
Proof.
  intros E a x H J; unfold fresh; exists E; intuition.
Qed.



Lemma swap_fresh_atoms :
  forall (a b : atom AR) (x : X),
  fresh a x -> fresh b x -> swap S (a, b) x = x.
Proof.
  unfold fresh; intros a b x H J.
  inversion_clear H as [E1 [H2 H3]]; inversion_clear J as [E2 [J2 J4]].
  assert (supports (intersection E1 E2) x); auto with nominal.
  rewrite swap_non_supports with (F := intersection E1 E2); auto with sets.
Qed.

End Support.

Implicit Arguments supports [AR X].
Implicit Arguments fresh [AR X].

Hint Resolve swap_non_supports : nominal.
Hint Resolve supports_subset : nominal.
Hint Resolve supports_intersection : nominal.
Hint Resolve supports_to_fresh : nominal.
Hint Resolve swap_fresh_atoms : nominal.
Hint Rewrite swap_fresh_atoms using (auto with nominal; fail) : nominal.


(* *********************************************************************** *)
(** ** Support for atoms *)

Lemma supports_atom :
  forall (A : AtomT) (a : atom A), supports (mkAtomSwap A) (singleton a) a.
Proof.
  unfold supports; unfold mkAtomSwap; unfold swap.
  intros A a b c B C.
  destruct_neg_singleton B.
  destruct_neg_singleton C.
  swapa_simpl; trivial.
Qed.

Hint Resolve supports_atom : nominal.


Lemma fresh_atom_from_neq :
  forall (A : AtomT) (a a' : atom A), a <> a' -> fresh (mkAtomSwap A) a a'.
Proof.
  unfold fresh; intros A a a' H; exists (@singleton _ (asetR A) a').
  split; auto with nominal sets.
Qed.

Hint Resolve fresh_atom_from_neq : nominal.


Lemma swap_not_in_set :
  forall (A : AtomT) (E : aset A) (a b c : atom A),
  ~ In a E ->
  ~ In b E ->
  ~ In c E ->
  ~ In (swap (mkAtomSwap A) (a, b) c) E.
Proof.
  intros A E a b c Q R S; simpl.
  atom_eqdec_case H a c;
    try atom_eqdec_case J b c; autorewrite with nominal; auto.
Qed.

Hint Resolve swap_not_in_set : nominal.


(* *********************************************************************** *)
(** ** Additional theorems. *)

Section GeneralTheorems.

Variable AR : AtomT.
Variable X : Set.
Variable XS : SwapT AR X.


(*
  XXX Historical note: These two lemmas were probably needed
  only during initial attempts at implementing a recursion operator.
*)
Lemma swap_fresh :
  forall a cd x, fresh XS a x -> fresh XS (swapa AR cd a) (swap XS cd x).
Proof.
  unfold fresh; intros a cd x H.
  inversion_clear H as [E [Q R]].
Admitted. (* XXX should be provable once extFsets are swap-sets *)


Lemma unswap_fresh :
  forall a cd x, fresh XS (swapa AR cd a) (swap XS cd x) -> fresh XS a x.
Proof.
  intros a cd x H.
  assert (J := swap_fresh _ cd _ H).
  rewrite swapa_invol in J; rewrite swap_invol in J; assumption.
Qed.

End GeneralTheorems.

Hint Resolve swap_fresh : nominal.
Hint Immediate unswap_fresh : nominal.


Section FunctionTheorems.

Variables AR : AtomT.
Variables X Y : Set.
Variables XS : SwapT AR X.
Variables YS : SwapT AR Y.


Lemma function_supports :
  forall (F : aset AR) (f : X -> Y),
  (forall a b x, ~ In a F -> ~ In b F ->
    swap YS (a, b) (f x) = f (swap XS (a, b) x)) ->
  supports (XS ^-> YS) F f.
Proof.
  unfold supports; intros F f H; intros c d C D.
  apply fun_extensionality; intros x; rewrite swap_func.
  rewrite <- H; trivial.
  rewrite swap_invol; reflexivity.
Qed.


Lemma application_supports :
  forall (F : aset AR) (f : X -> Y) (x : X),
  supports (XS ^-> YS) F f ->
  supports XS F x ->
  supports YS F (f x).
Proof.
  unfold supports; intros F f x H J a b A B.
  rewrite swap_app with (XS := XS).
  rewrite H; trivial.
  rewrite J; trivial.
Qed.


Lemma application_fresh :
  forall (a : AR) (f : X -> Y) (x : X),
  fresh (XS ^-> YS) a f -> fresh XS a x -> fresh YS a (f x).
Proof.
  unfold fresh; intros a f x H J.
  inversion_clear H as [E1 [H2 H3]].
  inversion_clear J as [E2 [J2 J3]].
  exists (union E1 E2); split.

    assert (supports (XS ^-> YS) (union E1 E2) f).
    eapply supports_subset; eauto with sets.
    assert (supports XS (union E1 E2) x).
    eapply supports_subset; eauto with sets.
    apply application_supports; trivial.

    auto with sets.
Qed.

End FunctionTheorems.


Section MajorTheorems.

Variables AR : AtomT.
Variables X : Set.
Variables XS : SwapT AR X.


Theorem function_constancy_v1 :
  forall (F : aset AR) (f : AR -> X),
  supports (mkAtomSwap AR ^-> XS) F f ->
  forall (a : AR), ~ In a F -> fresh XS a (f a) ->
  forall (b : AR), ~ In b F -> f a = f b.
Proof.
  intros F f H a' M1 M2 b Q.

  case (atom_eqdec AR b a'); intro L; try congruence.
  assert (fresh (mkAtomSwap AR) b a'); auto with nominal.
  assert (fresh (mkAtomSwap AR ^-> XS) b f); eauto with nominal.

  assert (fresh XS b (f a')).
  apply application_fresh with (XS := mkAtomSwap AR); trivial. 

  assert (Q1 : f a' = swap XS (b, a') (f a')).
  rewrite swap_fresh_atoms; trivial.

  generalize Q1.
  rewrite swap_app with (XS := mkAtomSwap AR).
  rewrite swap_non_supports with (F := F); trivial.
  simpl; autorewrite with nominal; congruence.
Qed.


Theorem function_constancy_v2 :
  forall (F : aset AR) (f : AR -> X),
  supports (mkAtomSwap AR ^-> XS) F f ->
  (exists a : AR, ~ In a F /\ fresh XS a (f a)) ->
  forall a b : AR, ~ In a F -> ~ In b F -> f a = f b.
Proof.
  intros F f H J a b A B.
  inversion_clear J as [a' [M1 M2]].
  rewrite <- (function_constancy_v1 F f H a' M1 M2 a); trivial.
  rewrite <- (function_constancy_v1 F f H a' M1 M2 b); trivial.
Qed.

End MajorTheorems.
